"""
preproc.py
==========

preprocessing functions (blinks, smoothing, missing data...)
"""
import numpy as np
import scipy.optimize
import pylab as plt

from .convenience import *


def smooth_window(x,window_len=11,window='hanning'):
    """smooth the data using a window with requested size.
    
    This method is based on the convolution of a scaled window with the signal.
    The signal is prepared by introducing reflected copies of the signal 
    (with the window size) in both ends so that transient parts are minimized
    in the begining and end part of the output signal.
    
    adapted from SciPy Cookbook: `<https://scipy-cookbook.readthedocs.io/items/SignalSmooth.html>`_.
    
    Parameters
    ----------    
    x: the input signal 
    window_len: the dimension of the smoothing window; should be an odd integer
    window: the type of window from 'flat', 'hanning', 'hamming', 'bartlett', 'blackman'
            flat window will produce a moving average smoothing.

    Returns
    -------
    np.array: the smoothed signal        
    """

    if x.ndim != 1:
        raise ValueError("smooth only accepts 1 dimension arrays.")

    if x.size < window_len:
        raise ValueError("Input vector needs to be bigger than window size.")


    if window_len<3:
        return x


    if not window in ['flat', 'hanning', 'hamming', 'bartlett', 'blackman']:
        raise ValueError("Window should be one of 'flat', 'hanning', 'hamming', 'bartlett', 'blackman'")


    s=np.r_[x[window_len-1:0:-1],x,x[-2:-window_len-1:-1]]
    if window == 'flat': #moving average
        w=np.ones(window_len,'d')
    else:
        w=eval('np.'+window+'(window_len)')

    y=np.convolve(w/w.sum(),s,mode='same')
    return y[(window_len-1):(-window_len+1)]


def detect_blinks_velocity(sy, smooth_winsize, vel_onset, vel_offset, min_onset_len=5, min_offset_len=5):
    """
    Detect blinks as everything between a fast downward and a fast upward-trending PD-changes.
    
    This works similarly to :py:func:`blink_onsets_mahot()`.
    
    Parameters
    ----------
    sy: np.array
        pupil data
    smooth_winsize: int (odd)
        size of the Hanning-window in sampling points
    vel_onset: float
        velocity-threshold to detect the onset of the blink
    vel_offset: float
        velocity-threshold to detect the offset of the blink
    min_onset_len: int
        minimum number of consecutive samples that cross the threshold to detect onset
    min_offset_len: int
        minimum number of consecutive samples that cross the threshold to detect offset
    """
    # generate smoothed signal and velocity-profile
    sym=smooth_window(sy, smooth_winsize, "hanning")
    vel=np.r_[0,np.diff(sym)] 
    n=sym.size

    # find first negative vel-crossing 
    onsets=np.where(vel<=vel_onset)[0]
    onsets_ixx=np.r_[np.diff(onsets),10]>1
    onsets_len=np.diff(np.r_[0,np.where(onsets_ixx)[0]])
    onsets=onsets[onsets_ixx]
    onsets=onsets[onsets_len>min_onset_len]

    ## offset finding
    offsets=np.where(vel>=vel_offset)[0]
    offsets_ixx=np.r_[10,np.diff(offsets)]>1
    offsets_len=np.diff(np.r_[np.where(offsets_ixx)[0],offsets.size])
    offsets=offsets[offsets_ixx]
    offsets=offsets[offsets_len>min_offset_len]
    
    
    ## find corresponding on- and off-sets
    blinks=[]
    on=onsets[0]
    while on is not None:
        offs=offsets[offsets>on]
        off=offs[0] if offs.size>0 else n
        blinks.append([on,off])
        ons=onsets[onsets>off]
        on=ons[0] if ons.size>0 else None
        
    ## if on- off-sets fall in a zero-region, grow until first non-zero sample
    blinks2=[]
    for (on,off) in blinks:
        while(on>0 and sy[on]==0):
            on-=1
        while(off<n-1 and sy[off]==0):
            off+=1
        blinks2.append([on,off])
    return np.array(blinks2)

def detect_blinks_zero(sy, min_duration, blink_val=0):
    """
    Detect blinks as consecutive sequence of `blink_val` (f.eks., 0 or NaN) of at least
    `min_duration` successive values in the signal `sy`.
    Detected blinks are put a matrix `blinks` (nblinks x 2) where start and end
    are stored as indexes.
    
    Parameters
    ----------
    sy: np.array (float)
        signal
    min_duration: int
        minimum number of consecutive samples for a sequence of missing numbers to be treated as blink
    blink_val: 
        "missing value" code
    
    Returns
    -------
    np.array (nblinks x 2) containing the indices of the start/end of the blinks
    """
    x=np.r_[0, np.diff((sy==blink_val).astype(np.int))]
    starts=np.where(x==1)[0]
    ends=np.where(x==-1)[0]-1
    if ends.size!=starts.size: 
        ## is the first start earlier than the first end?
        if starts[0]>ends[0]:
            ends=ends[1:] # drop first end
        else:
            starts=starts[:-1] # drop last start
    if ends[-1]==x.size:
        ends[-1]-=1
    blinks=[ [start,end] for start,end in zip(starts,ends) if end-start>=min_duration]
    return np.array(blinks)
    
    
def blink_onsets_mahot(sy, blinks, smooth_winsize, vel_onset, vel_offset, margin, blinkwindow):
    """
    Method for finding the on- and offset for each blink (excluding transient).
    See https://figshare.com/articles/A_simple_way_to_reconstruct_pupil_size_during_eye_blinks/688001.
    
    Parameters
    ----------
    sy: np.array
        pupil data
    blinks: np.array (nblinks x 2) 
        blink onset/offset matrix (contiguous zeros)
    smooth_winsize: int (odd)
        size of the Hanning-window in sampling points
    vel_onset: float
        velocity-threshold to detect the onset of the blink
    vel_offset: float
        velocity-threshold to detect the offset of the blink
    margin: tuple (int,int)
        margin that is subtracted/added to onset and offset (in sampling points)
    blinkwindow: int
        how much time before and after each blink to include (in sampling points)        
    """
    # generate smoothed signal and velocity-profile
    sym=smooth_window(sy, smooth_winsize, "hanning")
    vel=np.r_[0,np.diff(sym)] 
    blinkwindow_ix=blinkwindow
    n=sym.size
    
    newblinks=[]
    for ix,(start,end) in enumerate(blinks):                
        winstart,winend=max(0,start-blinkwindow_ix), min(end+blinkwindow_ix, n)
        slic=slice(winstart, winend) #start-blinkwindow_ix, end+blinkwindow_ix)
        winlength=vel[slic].size

        onsets=np.where(vel[slic]<=vel_onset)[0]
        offsets=np.where(vel[slic]>=vel_offset)[0]
        if onsets.size==0 or offsets.size==0:
            continue

        ## onsets are in "local" indices of the windows, start-end of blink global
        startl,endl=blinkwindow_ix if winstart>0 else start,end-start+blinkwindow_ix

        # find vel-crossing next to start of blink and move back to start of that crossing
        onset_ix=np.argmin(np.abs((onsets-startl<=0)*(onsets-startl)))
        while(onsets[onset_ix-1]+1==onsets[onset_ix]):
            onset_ix-=1
        onset=onsets[onset_ix]
        onset=max(0, onset-margin[0]) # avoid overflow to the left

        # find start of "reversal period" and move forward until it drops back
        offset_ix=np.argmin(np.abs(((offsets-endl<0)*np.iinfo(np.int).max)+(offsets-endl)))
        while(offset_ix<(len(offsets)-1) and offsets[offset_ix+1]-1==offsets[offset_ix]):
            offset_ix+=1        
        offset=offsets[offset_ix]
        offset=min(winlength-1, offset+margin[1]) # avoid overflow to the right
        newblinks.append( [onset+winstart,offset+winstart] )
    
    return np.array(newblinks)    


"""
convenience.py
==============

Some convenience functions.
"""


import numpy as np

def nprange(ar):
    return (ar.min(),ar.max())


import pystan
import pickle
from hashlib import md5

def zscale(y):
    return (y-np.nanmean(y))/np.nanstd(y)


def p_asym_laplac(y, mu, sigma, tau):
    """
    Asymmetric laplace distribution https://en.wikipedia.org/wiki/Asymmetric_Laplace_distribution;
    Parametrization as here: https://github.com/stan-dev/stan/issues/2312
    
    tau in [0,1]
    """
    I=np.array(y<=mu, dtype=np.int)
    return (2*tau*(1-tau))/sigma*np.exp(-2/sigma * ( (1-tau)*I*(mu-y) + tau*(1-I)*(y-mu) ) )

def p_asym_laplac_kappa(y, mu, lam, kappa):
    """
    Asymmetric laplace distribution https://en.wikipedia.org/wiki/Asymmetric_Laplace_distribution;
    Wikipedia parametrization.
    
    kappa in [0, infty] where 1 means symmetry
    """
    I=np.array(y<=mu, dtype=np.int)
    return (lam)/(kappa+1./kappa)*np.exp( ( (lam/kappa)*I*(y-mu) - lam*kappa*(1-I)*(y-mu) ) )


def trans_logistic_vec(x, a, b, inverse=False):
    """
    vectorized version of trans_logistic()
    
    goes from [a,b] to [-inf,+inf] and back;
    inverse=False: [a,b] -> [-inf, +inf]
    inverse=True:  [-inf,+inf] -> [a,b]
    if a or b is +/-infty, a logarithmic/exponential transform is used
    """
    eps=1e-15
    if inverse==False:
        # variables from [a,inf]
        x=np.where( (a>-np.infty) & (b==np.infty), np.log(np.maximum(x-a, eps)), x)
        # variables from [-inf, b]
        x=np.where( (a==-np.infty) & (b<np.infty), np.log(np.maximum(b-x, eps)), x)
        # variables from [a, b]
        x=np.where( (a>-np.infty) & (b<np.infty), -np.log( (b-a)/(x-a)-1 ), x)
    elif inverse==True:
        # variables from [-inf,inf] -> [a,inf]
        x=np.where( (a>-np.infty) & (b==np.infty), np.exp(x)+a, x)
        # variables from [-inf, inf] -> [-inf, b]
        x=np.where( (a==-np.infty) & (b<np.infty), b-np.exp(x), x)
        # variables from [-inf,inf] -> [a, b]
        x=np.where( (a>-np.infty) & (b<np.infty), (1./(1.+np.exp(-x)))*(b-a)+a, x)
    
    return x

def StanModel_cache(model_code, model_name=None, **kwargs):
    """Use just as you would `stan`"""
    code_hash = md5(model_code.encode('ascii')).hexdigest()
    if model_name is None:
        cache_fn = 'cached-model-{}.pkl'.format(code_hash)
    else:
        cache_fn = 'cached-{}-{}.pkl'.format(model_name, code_hash)
    try:
        sm = pickle.load(open(cache_fn, 'rb'))
    except:
        sm = pystan.StanModel(model_code=model_code)
        with open(cache_fn, 'wb') as f:
            pickle.dump(sm, f)
    else:
        print("Using cached StanModel")
    return sm



def plot_pupil_ipy(tx, sy, event_onsets=None, overlays=None, overlay_labels=None, 
                   blinks=None, interpolated=None,
                   figsize=(16,8), xlab="ms", nsteps=100):
    """
    Plotting with interactive adjustment of plotting window.
    To use this, do

    $ pip install ipywidgets
    $ jupyter nbextension enable --py widgetsnbextension
    $ jupyter labextension install @jupyter-widgets/jupyterlab-manager

    Parameters
    ----------
    
    tx : np.ndarray
        time-vector in seconds    
    sy : np.ndarray
        raw pupil signal        
    event_onsets : list
        onsets of events (stimuli/responses) in seconds
    overlays: tuple of np.array
        signals to overlay over the plot, given as tuple of arrays of same length as `tx`
    overlay_labels: tuple of strings
        labels for the overlays to be displayed in the legend
    figsize: tuple of int
        dimensions for the plot
    xlab: str
        label for x-axis
    nsteps: int
        number of steps for slider
    """
    import pylab as plt
    from ipywidgets import interact, interactive, fixed, interact_manual, Layout
    import ipywidgets as widgets

    def draw_plot(plotxrange):
        xmin,xmax=plotxrange
        ixmin=np.argmin(np.abs(tx-xmin))
        ixmax=np.argmin(np.abs(tx-xmax))
        plt.figure(figsize=figsize)

        plt.plot(tx[ixmin:ixmax],sy[ixmin:ixmax], label="signal")
        if overlays is not None:
            if type(overlays) is np.ndarray:
                plt.plot(tx[ixmin:ixmax],overlays[ixmin:ixmax],label=overlay_labels)
            else:
                for i,overlay in enumerate(overlays):
                    lab=overlay_labels[i] if overlay_labels is not None else None
                    plt.plot(tx[ixmin:ixmax],overlay[ixmin:ixmax], label=lab)
        for istart,iend in interpolated:
            plt.gca().axvspan(tx[istart],tx[iend],color="green", alpha=0.1)
        for istart,iend in blinks:
            plt.gca().axvspan(tx[istart],tx[iend],color="red", alpha=0.1)

        plt.vlines(event_onsets, *plt.ylim(), color="grey", alpha=0.5)
        plt.xlim(xmin,xmax)
        plt.xlabel(xlab)
        if overlay_labels is not None:
            plt.legend()


    wid_range=widgets.FloatRangeSlider(
        value=[tx.min(), tx.max()],
        min=tx.min(),
        max=tx.max(),
        step=(tx.max()-tx.min())/nsteps,
        description=' ',
        disabled=False,
        continuous_update=False,
        orientation='horizontal',
        readout=True,
        readout_format='.1f',
        layout=Layout(width='100%', height='80px')
    )

    interact(draw_plot, plotxrange=wid_range)
    

def helper_merge_blinks(b1,b2):
    if b1.size==0:
        return b2
    elif b2.size==0:
        return b1
    on=np.sort(np.concatenate( (b1[:,0], b2[:,0]) ))
    off=np.sort(np.concatenate( (b1[:,1], b2[:,1]) ))
    b=np.vstack((on,off)).T

    newb=[]
    on,off=b[0,:]
    for i in range(1,b.shape[0]):
        if b[i,0]<=off:
            # absorb onset from next 
            off=max(off,b[i,1])
        else:
            newb.append([on,off])
            on,off=b[i,:]
    off=b[-1,1]
    newb.append([on,off])
    return np.array(newb)


def sizeof_fmt(num, suffix='B'):
    """
    Convert number of bytes in `num` into human-readable string representation.
    Taken from https://stackoverflow.com/a/1094933
    """
    for unit in ['','Ki','Mi','Gi','Ti','Pi','Ei','Zi']:
        if abs(num) < 1024.0:
            return "%3.1f%s%s" % (num, unit, suffix)
        num /= 1024.0
    return "%.1f%s%s" % (num, 'Yi', suffix)


import unittest
import sys
import numpy as np
#sys.path.insert(0,"..")
#import pypillometry as pp
from .. import *

class TestPupilData(unittest.TestCase):
    def setUp(self):
        self.dfake=create_fake_pupildata(ntrials=100, fs=500)
        self.d=PupilData.from_file("data/test.pd")
    def test_from_file(self):
        d=PupilData.from_file("data/test.pd")
        self.assertEqual(d.__class__, PupilData)
    def test_create_fakedata(self):
        d=create_fake_pupildata(ntrials=100)
        self.assertEqual(d.__class__, FakePupilData)
    def test_history(self):
        d=create_fake_pupildata(ntrials=100)
        self.assertEqual(len(d.history), 0)
        d=d.drop_original()
        self.assertEqual(len(d.history), 1)
    def test_drop_original(self):
        d2=self.dfake.drop_original()
        self.assertIsNone(d2.original)
        self.assertLess(d2.size_bytes(), self.dfake.size_bytes())
    def test_reset_time(self):
        d=self.dfake.reset_time(t0=500)
        self.assertEqual(d.tx[0], 500)
        d=d.reset_time()
        self.assertEqual(d.tx[0], 0)
    def test_summary(self):
        self.dfake.summary()
        self.d.summary()
    def test_len(self):
        self.assertEqual(self.d.sy.size, len(self.d))
    def test_nevents(self):
        self.d.nevents()
    def test_nblinks(self):
        self.assertEqual(self.d.nblinks(), 0)
        d=self.d.blinks_detect()
        self.assertGreater(d.nblinks(), 0)
    def test_get_duration(self):
        d1=self.d.get_duration(units="ms")
        d2=self.d.get_duration(units="sec")
        d3=self.d.get_duration(units="min")
        d4=self.d.get_duration(units="h")
        self.assertLess(d2,d1)
        self.assertLess(d3,d2)
        self.assertLess(d4,d3)
        self.assertAlmostEqual(d2, d1/1000.)
        self.assertAlmostEqual(d3, d2/60.)
        self.assertAlmostEqual(d4, d3/60.)
    def test_sub_slice(self):
        d=self.dfake.sub_slice(1, 2, units="min")
        diff=d.tx[1]-d.tx[0]
        self.assertLess(abs(d.tx[0]-1*60*1000.), diff)
        self.assertLess(abs(d.get_duration(units="min")-1), diff)
    def test_scale(self):
        d=self.dfake.scale()
        self.assertAlmostEqual(np.mean(d.sy), 0)
        self.assertAlmostEqual(np.std(d.sy), 1)
        d2=d.unscale()
        self.assertAlmostEqual(np.mean(d2.sy), np.mean(self.dfake.sy))
        self.assertAlmostEqual(np.std(d2.sy), np.std(self.dfake.sy))
    def test_lowpass_filter(self):
        d=self.dfake.lowpass_filter(2)
    def test_smooth_window(self):
        d=self.dfake.smooth_window()
        d2=d.smooth_window(window="bartlett")
        d3=d.smooth_window(winsize=20)
    def test_downsample(self):
        d=self.dfake.downsample(100)
        self.assertEqual(d.fs, 100)
        with self.assertRaises(ZeroDivisionError):
            d.downsample(101)
        d2=self.dfake.downsample(5, dsfac=True)
        self.assertAlmostEqual(d.fs, d2.fs)
    def test_copy(self):
        d=self.dfake.copy()
        d.sy[0]=self.dfake.sy[0]+1.0
        self.assertNotEqual(d.sy[0], self.dfake.sy[0])
    def test_estimate_baseline(self):
        d=self.dfake.estimate_baseline()
        self.assertLess(np.sum(d.sy<d.baseline), 0.1*len(d))
    def test_stat_per_event(self):
        a1=self.dfake.stat_per_event([-100,0], return_missing=None)
        a2=self.dfake.stat_per_event([-100,0], return_missing="nmiss")
        a3=self.dfake.stat_per_event([-100,0], return_missing="prop")
        self.assertEqual(a1.__class__, np.ndarray)
        self.assertEqual(a2.__class__, tuple)
        self.assertEqual(a3.__class__, tuple)
    def test_estimate_response(self):
        d=self.dfake.estimate_response()
    def test_blinks_detect(self):
        d=self.d.blinks_detect()
    def test_blinks_merge(self):
        d=self.d.blinks_detect()
        d2=d.blinks_merge()
    def test_blinks_merge(self):
        d=self.d.blinks_detect().blinks_merge().blinks_interpolate()
    def test_blinks_merge(self):
        d=self.d.blinks_detect().blinks_merge().blinks_interp_mahot()
    def test_get_erpd(self):
        a=self.d.get_erpd("test", lambda x: True)
        self.assertEqual(a.__class__, ERPD)
    
    
    
        
        
    
        

if __name__ == '__main__':
    unittest.main()


"""
pupildata.py
============

Main object-oriented entry point
"""

from .convenience import *
from .baseline import *
from .fakedata import *
from .preproc import *
from .io import *
from .erpd import *

import pylab as plt
import matplotlib as mpl
from matplotlib.backends.backend_pdf import PdfPages
import numpy as np
import scipy.signal as signal
from scipy.interpolate import interp1d
from scipy import interpolate
import scipy
from random import choice
import pickle

import collections.abc

import copy
import math

#from pytypes import typechecked
from typing import Sequence, Union, List, TypeVar, Optional, Tuple, Callable
PupilArray=Union[np.ndarray, List[float]]


_inplace=False ## default for whether or not inplace-operations should be used


import inspect
import functools

## decoratory to keep a history of actions performed on dataset
# can only be used with functions that return "self" 
def keephistory(func):
    @functools.wraps(func)
    def wrapper(*args,**kwargs):
        obj=func(*args,**kwargs)
        funcname=func.__name__
        argstr=",".join(["%s"%(v) for v in args[1:]])
        kwargstr=",".join(["%s=%s"%(k,v) for k,v in kwargs.items()])
        allargs=argstr
        if len(allargs)>0 and len(kwargstr)>0:
            allargs+=","+kwargstr
        elif len(kwargstr)>0:
            allargs+=kwargstr
        fstr="{func}({allargs})".format(func=funcname, allargs=allargs)
        #fstr="{func}({argstr},{kwargstr})".format(func=funcname,argstr=argstr,kwargstr=kwargstr)
        obj.add_to_history({"funcstring":fstr, "funcname":funcname, "args":args[1:], "kwargs":kwargs})
        return obj
    return wrapper
        
             

#@typechecked
class PupilData:
    """
    Class representing pupillometric data. 
    """
    
    def add_to_history(self, event):
        """Add event to history"""
        try:
            self.history.append(event)
        except:
            self.history=[event]
            
    def print_history(self):
        """
        Pretty-print the history of the current dataset (which manipulations where done on it).
        """
        print("* "+self.name)
        try:
            for i,ev in enumerate(self.history):
                print(" "*(i)+"└ " + ev["funcstring"])
        except:
            print("no history")
   
    def apply_history(self, obj):
        """
        Apply history of operations done on `self` to `obj`.
        
        Parameters:
        -----------
        
        obj: :class:`.PupilData`
            object of class :class:`.PupilData` to which the operations are to be transferred
            
        Returns:
        --------
        
        copy of the :class:`.PupilData`-object to which the operations in `self` were applied
        """
        for ev in self.history:
            obj=getattr(obj, ev["funcname"])(*ev["args"], **ev["kwargs"])
        return obj

    def __len__(self) -> int:
        """Return number of sampling points in the pupil data."""
        return self.sy.size
    
    def nevents(self) -> int:
        """Return number of events in pupillometric data."""
        return self.event_onsets.size

    def nblinks(self) -> int:
        """
        Return number of detected blinks. Should be run after `detect_blinks()`.
        """
        return self.blinks.shape[0]
    
    def get_duration(self, units="min"):
        fac=self._unit_fac(units)
        return (len(self)/self.fs*1000)*fac
    
    def _random_id(self, n:int=8) -> str:
        """
        Create a random ID string that is easy to recognise.
        Based on <http://code.activestate.com/recipes/526619-friendly-readable-id-strings/>.
        """
        v = 'aeiou'
        c = 'bdfghklmnprstvw'

        return ''.join([choice(v if i%2 else c) for i in range(n)])
    
    def __init__(self,
                 pupil: PupilArray, 
                 sampling_rate: Optional[float]=None,
                 time: Optional[PupilArray]=None,
                 event_onsets: Optional[PupilArray]=None,
                 event_labels: Optional[PupilArray]=None,
                 name: Optional[str]=None,
                 keep_orig: bool=True,
                 fill_time_discontinuities: bool=True):
        """
        Parameters
        ----------
        
        name: 
            name of the dataset or `None` (in which case a random string is selected)
        time: 
            timing array or `None`, in which case the time-array goes from [0,maxT]
            using `sampling_rate` (in ms)
        pupil:
            pupillary data at times `time` assumed to be in ms
        event_onsets: 
            time-onsets of any events that are to be modelled in the pupil
        event_labels:
            for each event in `event_onsets`, provide a label
        sampling_rate: float
            sampling-rate of the pupillary signal in Hz
        keep_orig: bool
            keep a copy of the original dataset? If `True`, a copy of the :class:`.PupilData` object
            as initiated in the constructor is stored in member `PupilData.original`
        fill_time_discontinuities: bool
            sometimes, when the eyetracker loses signal, no entry in the EDF is made; 
            when this option is True, such entries will be made and the signal set to 0 there
        """
        self.sy=np.array(pupil, dtype=np.float)
        if sampling_rate is None and time is None:
            raise ValueError("you have to specify either sampling_rate or time-vector (or both)")
        
        if time is None:
            maxT=len(self)/sampling_rate*1000.
            self.tx=np.linspace(0,maxT, num=len(self))
        else:
            self.tx=np.array(time, dtype=np.float)
        
        if sampling_rate is None:
            self.fs=np.round(1000./np.median(np.diff(self.tx)))
        else:
            self.fs=sampling_rate
            
        if fill_time_discontinuities:
            ## find gaps in the time-vector
            tx=self.tx
            sy=self.sy
            stepsize=np.median(np.diff(tx))
            n=tx.size
            gaps_end_ix=np.where(np.r_[stepsize,np.diff(tx)]>2*stepsize)[0]
            ngaps=gaps_end_ix.size
            if ngaps!=0:
                ## at least one gap here
                print("> Filling in %i gaps"%ngaps)
                gaps_start_ix=gaps_end_ix-1
                print( ((tx[gaps_end_ix]-tx[gaps_start_ix])/1000), "seconds" )
                
                ntx=[tx[0:gaps_start_ix[0]]] # initial
                nsy=[sy[0:gaps_start_ix[0]]]
                for i in range(ngaps):
                    start,end=gaps_start_ix[i], gaps_end_ix[i]
                    # fill in the gap
                    ntx.append( np.linspace(tx[start],tx[end], int((tx[end]-tx[start])/stepsize), endpoint=False) )
                    nsy.append( np.zeros(ntx[-1].size) )

                    # append valid signal
                    if i==ngaps-1:
                        nstart=n
                    else:
                        nstart=gaps_start_ix[i+1]
                    ntx.append( tx[end:nstart] )
                    nsy.append( sy[end:nstart] )

                ntx=np.concatenate(ntx)
                nsy=np.concatenate(nsy) 
                self.tx=ntx
                self.sy=nsy
            
        if event_onsets is None:
            self.event_onsets=np.array([], dtype=np.float)
        else:
            self.event_onsets=np.array(event_onsets, dtype=np.float)
        
        # check whether onsets are in range
        for onset in self.event_onsets:
            if onset<self.tx.min() or onset>self.tx.max():
                raise ValueError("some event-onsets outside data range according to time-vector")
            
            
        if event_labels is None:
            self.event_labels=np.zeros_like(self.event_onsets)
        else:
            if self.event_onsets.size!=np.array(event_labels).size:
                raise ValueError("event_labels must have same length as event_onsets")
            self.event_labels=np.array(event_labels)
          
        if self.tx.size != self.sy.size:
            raise ValueError("time and pupil-array must have same length, found {} vs {}".format(
                self.tx.size,self.sy.size))
        
        if name is None:
            self.name = self._random_id()
        else:
            self.name=name
        
        ## initialize baseline signal
        self.scale_params={"mean":0, "sd":1}
        self.baseline=np.zeros(len(self))
        self.baseline_estimated=False
        
        ## initialize response-signal
        self.response=np.zeros(len(self))
        self.response_pars=None
        self.response_estimated=False
        
        ## initialize blinks
        self.blinks=np.empty((0,2), dtype=np.int)
        self.blink_mask=np.zeros(len(self), dtype=np.int)
        
        ## interpolated mask
        self.interpolated_mask=np.zeros(len(self), dtype=np.int)
        self.missing=np.zeros(len(self), dtype=np.int)
        self.missing[self.sy==0]=1
        
        self.original=None
        if keep_orig: 
            self.original=self.copy()
            
        ## start with empty history    
        self.history=[]
       
    @keephistory
    def drop_original(self, inplace=_inplace):
        """
        Drop original dataset from record (e.g., to save space).
        """
        obj=self if inplace else self.copy()
        obj.original=None
        return obj
    
    @keephistory
    def reset_time(self, t0: float=0, inplace=_inplace):
        """
        Resets time so that the time-array starts at time zero (t0).
        Resets onsets etc.
        
        Parameters
        ----------
        t0: float
            time at which the :class:`.PupilData`'s time-vector starts
        inplace: bool
            if `True`, make change in-place and return the object
            if `False`, make and return copy before making changes
        """
        tmin=self.tx.min()
        obj=self if inplace else self.copy()            
        obj.tx=(self.tx-tmin)+t0
        obj.event_onsets=(self.event_onsets-tmin)+t0
        return obj
        
    def write_file(self, fname:str):
        """
        Save to file (using :mod:`pickle`).
        
        Parameters
        ----------
        
        fname: str
            filename
        """
        pd_write_pickle(self, fname)
       
    @classmethod
    def from_file(cls, fname:str):
        """
        Reads a :class:`.PupilData` object from a pickle-file.
        Use as ``pypillometry.PupilData.from_file("yourfile.pd")``.
        
        Parameters
        ----------
        
        fname: str
            filename
        """
        r=pd_read_pickle(fname)
        return r
        
    def _unit_fac(self, units):
        if units=="sec":
            fac=1./1000.
        elif units=="min":
            fac=1./1000./60.
        elif units=="h":
            fac=1./1000./60./60.
        else:
            fac=1.
        return fac

    @keephistory
    def sub_slice(self, start: float=-np.inf, end: float=np.inf, units: str="sec"):
        """
        Return a new `PupilData` object that is a shortened version
        of the current one (contains all data between `start` and
        `end` in units given by `units` (one of "ms", "sec", "min", "h").

        Parameters
        ----------
        
        start: float
            start for new dataset
        end: float
            end of new dataset
        units: str
            time units in which `start` and `end` are provided
        """
        slic=self.copy()
        fac=self._unit_fac(units)
        tx=self.tx*fac
        keepix=np.where(np.logical_and(tx>=start, tx<=end))
        for k, v in slic.__dict__.items():
            if isinstance(v,np.ndarray) and v.size==self.sy.size:
                slic.__dict__[k]=slic.__dict__[k][keepix]
        evon=slic.event_onsets*slic._unit_fac(units)
        keepev=np.logical_and(evon>=start, evon<=end)
        slic.event_onsets=slic.event_onsets[keepev]
        slic.event_labels=slic.event_labels[keepev]
        ## just remove all detected blinks (need to rerun `detect_blinks()`)
        slic.blinks=np.empty((0,2), dtype=np.int)
        slic.blink_mask=np.zeros(len(slic), dtype=np.int)
        return slic

    def summary(self) -> dict:
        """Return a summary of the :class:`.PupilData`-object."""
        summary=dict(
            name=self.name,
            n=len(self),
            nmiss=np.sum(self.missing),#np.sum(np.isnan(self.sy))+np.sum(self.sy==0),
            perc_miss=np.sum(self.missing)/len(self)*100.,#(np.sum(np.isnan(self.sy))+np.sum(self.sy==0))/len(self)*100.,
            nevents=self.nevents(), 
            nblinks=self.nblinks(),
            ninterpolated=self.interpolated_mask.sum(),
            blinks_per_min=self.nblinks()/(len(self)/self.fs/60.),
            fs=self.fs, 
            duration_minutes=self.get_duration("min"),
            start_min=self.tx.min()/1000./60.,
            end_min=self.tx.max()/1000./60.,
            baseline_estimated=self.baseline_estimated,
            response_estimated=self.response_estimated)        
        return summary
    
    def size_bytes(self):
        """
        Return size of current dataset in bytes.
        """
        nbytes=len(pickle.dumps(self, -1))
        return nbytes
    
    def __repr__(self) -> str:
        """Return a string-representation of the dataset."""
        pars=self.summary()
        del pars["name"]
        s="PupilData({name}, {size}):\n".format(name=self.name, size=sizeof_fmt(self.size_bytes()))
        flen=max([len(k) for k in pars.keys()])
        for k,v in pars.items():
            s+=(" {k:<"+str(flen)+"}: {v}\n").format(k=k,v=v)
        s+=" History:\n *\n"
        try:
            for i,ev in enumerate(self.history):
                s+=" "*(i+1)+"└ " + ev["funcstring"] +"\n"
        except:
            s+=" └no history\n"
        return s
    
    @keephistory    
    def unscale(self, mean: Optional[float]=None, sd: Optional[float]=None, inplace=_inplace):
        """
        Scale back to original values using either values provided as arguments
        or the values stored in `scale_params`.
        
        Parameters
        ----------
        mean: float
            mean to add from signal
        sd: float
            sd to scale with        
        inplace: bool
            if `True`, make change in-place and return the object
            if `False`, make and return copy before making changes
        
        """
        if mean is None:
            mean=self.scale_params["mean"]
        if sd is None:
            sd=self.scale_params["sd"]
        
        obj=self if inplace else self.copy()
        obj.scale_params={"mean":0, "sd":1}
        obj.sy=(self.sy*sd)+mean
        obj.baseline=(self.baseline*sd)+mean
        obj.response=(self.response*sd)
        return obj
    
    @keephistory
    def scale(self, mean: Optional[float]=None, sd: Optional[float]=None, inplace=_inplace):
        """
        Scale the pupillary signal by subtracting `mean` and dividing by `sd`.
        If these variables are not provided, use the signal's mean and std.
        
        Parameters
        ----------
        
        mean: float
            mean to subtract from signal
        sd: float
            sd to scale with
        inplace: bool
            if `True`, make change in-place and return the object
            if `False`, make and return copy before making changes        
        
        Note
        ----
        Scaling-parameters are being saved in the `scale_params` argument. 
        """
        if mean is None:
            mean=np.nanmean(self.sy)
        if sd is None:
            sd=np.nanstd(self.sy)

        obj=self if inplace else self.copy()            
        obj.scale_params={"mean":mean, "sd":sd}
        obj.sy=(self.sy-mean)/sd
        obj.baseline=(self.baseline-mean)/sd
        obj.response=(self.response)/sd
        return obj
    
    @keephistory
    def lowpass_filter(self, cutoff: float, order: int=2, inplace=_inplace):
        """
        Lowpass-filter signal using a Butterworth-filter, 
        see :py:func:`pypillometry.baseline.butter_lowpass_filter()`.
    
        Parameters
        -----------

        cutoff: float
            lowpass-filter cutoff
        order: int
            filter order
        inplace: bool
            if `True`, make change in-place and return the object
            if `False`, make and return copy before making changes                    
        """
        obj=self if inplace else self.copy()                    
        obj.sy=butter_lowpass_filter(self.sy, cutoff, self.fs, order)
        return obj

    @keephistory
    def smooth_window(self, window: str="hanning", winsize: float=11, inplace=_inplace):
        """
        Apply smoothing of the signal using a moving window. See :func:`.smooth_window()`.
        
        Parameters
        ----------
        window: str
            (the type of window from 'flat', 'hanning', 'hamming', 'bartlett', 'blackman'); 
             flat window will produce a moving average smoothing.
        winsize: float
            the length of the window in ms 
        inplace: bool
            if `True`, make change in-place and return the object
            if `False`, make and return copy before making changes                            
        """
        winsize_ix=int(np.ceil(winsize/1000.*self.fs)) 
        obj=self if inplace else self.copy()                            
        obj.sy=smooth_window(self.sy, winsize_ix, window )
        return obj
    
    @keephistory
    def downsample(self, fsd: float, dsfac: bool=False, inplace=_inplace):
        """
        Simple downsampling scheme using mean within the downsampling window.
        See :py:func:`pypillometry.baseline.downsample()`.

        Parameters
        -----------

        fsd: 
            new sampling-rate or decimate-factor
        dsfac:
            if False, `fsd` is the new sampling rate;
            if True, `fsd` is the decimate factor
        inplace: bool
            if `True`, make change in-place and return the object
            if `False`, make and return copy before making changes                                        
        """
        if dsfac:
            dsfac=fsd
            fsd=float(self.fs/dsfac)
        else:
            dsfac=int(self.fs/fsd) # calculate downsampling factor
            
        obj=self if inplace else self.copy()
        
        ## downsample all arrays that have the original sy-length
        # (this is so that the function is general for subclasses, as well)
        nd=self.sy.size
        for k, v in obj.__dict__.items():
            if isinstance(v,np.ndarray) and v.size==nd:
                obj.__dict__[k]=downsample(self.__dict__[k], dsfac)
            
        obj.fs=fsd
        return obj

    def copy(self, new_name: Optional[str]=None):
        """
        Make and return a deep-copy of the pupil data.
        """
        cls = self.__class__
        result = cls.__new__(cls)
        for k, v in self.__dict__.items():
            setattr(result, k, copy.deepcopy(v))
        if new_name is None:
            result.name=self.name+"_"+self._random_id(n=2)
        else:
            result.name=new_name
        return result        

    def _plot(self, plot_range, overlays, overlay_labels, units, interactive, highlight_blinks, highlight_interpolated):
        fac=self._unit_fac(units)
        if units=="sec":
            xlab="seconds"
        elif units=="min":
            xlab="minutes"
        elif units=="h":
            xlab="hours"
        else:
            xlab="ms"
        tx=self.tx*fac
        evon=self.event_onsets*fac
        
        start,end=plot_range
        if start==-np.infty:
            startix=0
        else:
            startix=np.argmin(np.abs(tx-start))
            
        if end==np.infty:
            endix=tx.size
        else:
            endix=np.argmin(np.abs(tx-end))
        
        tx=tx[startix:endix]
        
        ixx=np.logical_and(evon>=start, evon<end)
        evlab=self.event_labels[ixx]
        evon=evon[ixx]
        overlays=(ov[startix:endix] for ov in overlays)
        
        if interactive:
            blinks=np.empty((0,2), dtype=np.int)
            interpolated=np.empty((0,2), dtype=np.int)
            if highlight_blinks:
                blinks=[]
                for sblink,eblink in self.blinks:
                    if eblink<startix or sblink>endix:
                        continue
                    else:
                        sblink=max(0,sblink-startix)
                        eblink=min(endix,eblink-startix)
                    blinks.append([sblink,eblink])
                blinks=np.array(blinks)
            if highlight_interpolated:
                a=np.diff(np.r_[0, self.interpolated_mask[startix:endix], 0])[:-1]
                istarts=np.where(a>0)[0]
                iends=np.where(a<0)[0]
                interpolated=[]
                for istart,iend in zip(istarts,iends):
                    interpolated.append([istart,iend])
            plot_pupil_ipy(tx, self.sy[startix:endix], evon,
                           overlays=overlays, overlay_labels=overlay_labels,
                           blinks=blinks, interpolated=interpolated,
                          xlab=xlab)
        else:
            plt.plot(tx, self.sy[startix:endix], label="signal")
            for i,ov in enumerate(overlays):
                plt.plot(tx, ov, label=overlay_labels[i])
            plt.vlines(evon, *plt.ylim(), color="grey", alpha=0.5)
            ll,ul=plt.ylim()
            for ev,lab in zip(evon,evlab):
                plt.text(ev, ll+(ul-ll)/2., "%s"%lab, fontsize=8, rotation=90)
            if highlight_interpolated:
                a=np.diff(np.r_[0, self.interpolated_mask[startix:endix], 0])[:-1]
                istarts=np.where(a>0)[0]
                iends=np.where(a<0)[0]
                for istart,iend in zip(istarts,iends):
                    plt.gca().axvspan(tx[istart],tx[iend],color="green", alpha=0.1)
            if highlight_blinks:
                for sblink,eblink in self.blinks:
                    if eblink<startix or sblink>endix:
                        continue
                    else:
                        sblink=min(tx.size-1, max(0,sblink-startix))
                        eblink=min(endix-startix-1,eblink-startix)
                    
                    plt.gca().axvspan(tx[sblink],tx[eblink],color="red", alpha=0.2)
                
                
            plt.legend()
            plt.xlabel(xlab)        
    
    def plot(self, plot_range: Tuple[float,float]=(-np.infty, +np.infty),
             interactive: bool=False, 
             baseline: bool=True, 
             response: bool=False,
             model: bool=True,
             highlight_blinks: bool=True,
             highlight_interpolated: bool=True,
             units: str="sec"
            ) -> None:
        """
        Make a plot of the pupil data using `matplotlib` or :py:func:`pypillometry.convenience.plot_pupil_ipy()`
        if `interactive=True`.

        Parameters
        ----------
        plot_range: tuple (start,end)
            plot from start to end (in units of `units`)
        baseline: bool
            plot baseline if estimated
        response: bool
            plot response if estimated
        model: bool
            plot full model if baseline and response have been estimated
        interactive: bool
            if True, plot with sliders to adjust range
        units: str
            one of "sec"=seconds, "ms"=millisec, "min"=minutes, "h"=hours
        """

        overlays=tuple()
        overlay_labels=tuple()
        if baseline and self.baseline_estimated:
            overlays+=(self.baseline,)                
            overlay_labels+=("baseline",)
        if response and self.response_estimated:
            overlays+=(self.response,)
            overlay_labels+=("response",)             
        if model and self.baseline_estimated and self.response_estimated:
            overlays+=(self.baseline+self.response,)
            overlay_labels+=("model",)
        self._plot(plot_range, overlays, overlay_labels, units, interactive, highlight_blinks, highlight_interpolated)

    def plot_segments(self, overlay=None, pdffile: Optional[str]=None, interv: float=1, figsize=(15,5), ylim=None, **kwargs):
        """
        Plot the whole dataset chunked up into segments (usually to a PDF file).

        Parameters
        ----------

        pdffile: str or None
            file name to store the PDF; if None, no PDF is written 
        interv: float
            duration of each of the segments to be plotted (in minutes)
        figsize: Tuple[int,int]
            dimensions of the figures
        kwargs: 
            arguments passed to :func:`.PupilData.plot()`

        Returns
        -------

        figs: list of :class:`matplotlib.Figure` objects
        """

        # start and end in minutes
        smins,emins=self.tx.min()/1000./60., self.tx.max()/1000./60.
        segments=[]
        cstart=smins
        cend=smins
        while cend<emins:
            cend=min(emins, cstart+interv)
            segments.append( (cstart,cend) )
            cstart=cend

        figs=[]
        _backend=mpl.get_backend()
        mpl.use("pdf")
        plt.ioff() ## avoid showing plots when saving to PDF 

        for start,end in segments:
            plt.figure(figsize=figsize)
            self.plot( (start,end), units="min", **kwargs)
            if overlay is not None:
                overlay.plot( (start, end), units="min", **kwargs)  
            if ylim is not None:
                plt.ylim(*ylim)
            figs.append(plt.gcf())


        if isinstance(pdffile, str):
            print("> Writing PDF file '%s'"%pdffile)
            with PdfPages(pdffile) as pdf:
                for fig in figs:
                    pdf.savefig(fig)         

        ## switch back to original backend and interactive mode                        
        mpl.use(_backend) 
        plt.ion()

        return figs        
    
    @keephistory
    def estimate_baseline(self, method: str="envelope_iter_bspline_2", inplace=_inplace, **kwargs):
        """
        Apply one of the baseline-estimation methods.
        
        Parameters
        ----------
        
        method: 
            "envelope_iter_bspline_1": :py:func:`pypillometry.baseline.baseline_envelope_iter_bspline()` 
                                        with one iteration
            "envelope_iter_bspline_2": :py:func:`pypillometry.baseline.baseline_envelope_iter_bspline()` 
                                        with two iterations
        inplace: bool
            if `True`, make change in-place and return the object
            if `False`, make and return copy before making changes                                        
            
        kwargs:
            named arguments passed to the low-level function in :py:mod:`pypillometry.baseline`.
            
        Note
        -----
        the results of the estimation is stored in member `baseline`
        
        """
        obj=self if inplace else self.copy()
        if method=="envelope_iter_bspline_2":
            txd,syd,base2,base1=baseline_envelope_iter_bspline(self.tx, self.sy,self.event_onsets,self.fs,**kwargs)
            f=interpolate.interp1d(txd, base2, kind="cubic", bounds_error=False, fill_value="extrapolate")
            obj.baseline=f(self.tx)
        elif method=="envelope_iter_bspline_1": 
            txd,syd,base2,base1=baseline_envelope_iter_bspline(self.tx, self.sy,self.event_onsets,self.fs,**kwargs)
            f=interpolate.interp1d(txd, base1, kind="cubic", bounds_error=False, fill_value="extrapolate")
            obj.baseline=f(self.tx)            
        else:
            raise ValueError("Undefined method for baseline estimation: %s"%method)         
        obj.baseline_estimated=True
        return obj

    def stat_per_event(self, interval: Tuple[float,float], event_select=None, statfct: Callable=np.mean, return_missing: Optional[str]=None):
        """
        Return result of applying a statistical function to pupillometric data in a
        given interval relative to event-onsets. For example, extract mean 
        pupil-size in interval before trial onset.

        Parameters
        -----------
        event_select: str or function
            variable describing which events to select and align to
            - if str: use all events whose label contains the string
            - if function: apply function to all labels, use those where the function returns True
        
        interval : tuple (min,max)
            time-window in ms relative to event-onset (0 is event-onset)

        statfct : function
            function mapping np.array to a single number

        return_missing: None, "nmiss", "prop"
            if None, only an array with the stats per event is return
            if "nmiss", returns a tuple (stat, nmiss) where `nmiss` is the number of missing vales in the timewin
            if "prop", return a tuple (stat, prop_miss) where `prop_miss` is the proportion missing vales in the timewin
    
        Returns
        --------

        result: np.array
            number of event-onsets long result array
        """
        if callable(event_select):
            event_ix=np.array([bool(event_select(evlab)) for evlab in self.event_labels])
        elif isinstance(event_select, str):
            event_ix=np.array([event_select in evlab for evlab in self.event_labels])
        elif event_select is None:
            event_ix=np.arange(self.nevents())
        else:
            raise ValueError("event_select must be string or function")
        
        stat =stat_event_interval(self.tx, self.sy, self.event_onsets[event_ix], interval, statfct)
        if return_missing=="nmiss":
            nmiss=stat_event_interval(self.tx, np.logical_or(self.missing, self.interpolated_mask), 
                                      self.event_onsets[event_ix], interval, np.sum)
            ret=(stat,nmiss)
        elif return_missing=="prop":
            prop_miss=stat_event_interval(self.tx, np.logical_or(self.missing, self.interpolated_mask), 
                                          self.event_onsets[event_ix], interval, np.mean)
            ret=(stat,prop_miss)            
        else: 
            ret=stat
        return ret
    
    @keephistory
    def estimate_response(self, npar: Union[str,float]="free", tmax: Union[str,float]="free", 
                          verbose: int=50,
                          bounds: dict={"npar":(1,20), "tmax":(100,2000)},
                          inplace=_inplace):
        """
        Estimate pupil-response based on event-onsets, see
        :py:func:`pypillometry.pupil.pupil_response()`.
        

        npar: float
            npar-parameter for the canonical response-function or "free";
            in case of "free", the function optimizes for this parameter
        tmax: float
            tmax-parameter for the canonical response-function or "free";
            in case of "free", the function optimizes for this parameter
        bounds: dict
            in case that one or both parameters are estimated, give the lower
            and upper bounds for the parameters        
        inplace: bool
            if `True`, make change in-place and return the object
            if `False`, make and return copy before making changes                                        
        
        Note
        ----
        the results of the estimation is stored in members `response`, `response_x` (design matrix) 
        and `response_pars`

        """
        if not self.baseline_estimated:
            print("WARNING: no baseline estimated yet, using zero as baseline")
        
        pred, coef, npar_est, tmax_est, x1=pupil_response(self.tx, self.sy-self.baseline, 
                                                          self.event_onsets, self.fs, 
                                                          npar=npar, tmax=tmax, verbose=verbose,
                                                         bounds=bounds)
        obj=self if inplace else self.copy()
        obj.response_pars={"npar":npar_est,
                            "npar_free":True if npar=="free" else False,
                            "tmax":tmax_est,
                            "tmax_free":True if tmax=="free" else False,
                            "coef":coef,
                            "bounds":bounds
                           }
        
        obj.response=pred
        obj.response_x=x1
        obj.response_estimated=True
        return obj
    

    @keephistory
    def blinks_detect(self, min_duration:float=20, blink_val:float=0,
                      winsize: float=11, vel_onset: float=-5, vel_offset: float=5, 
                      min_onset_len: int=5, min_offset_len: int=5,
                      strategies: List[str]=["zero","velocity"],
                      units="ms", inplace=_inplace):
        """
        Detect blinks in the pupillary signal using several strategies.
        First, blinks are detected as consecutive sequence of `blink_val` 
        (f.eks., 0 or NaN). Second, blinks are defined as everything between
        two crossings of the velocity profile (from negative to positive).
        
        Detected blinks are put into member `blinks` (matrix 2 x nblinks where start and end
        are stored as indexes) and member `blink_mask` which codes for each sampling point
        whether there is a blink (1) or not (0).

        Finally, detected blinks have to be at least `min_duration` duration (in `units`).
        
        Parameters
        ----------
        min_duration: float
            minimum duration for a sequence of missing numbers to be treated as blink
        blink_val: float
            "missing value" code
        winsize:
            window-size for smoothing for velocity profile (in units)
        vel_onset:
            negative velocity that needs to be crossed; arbitrary units that depend on
            sampling rate etc
        vel_offset:
            positive velocity that needs to be exceeded; arbitrary units that depend on
            sampling rate etc
        min_onset_len: int
            minimum number of consecutive samples that crossed threshold in the velocity
            profile to detect as onset (to avoid noise-induced changes)
        min_offset_len: int
            minimum number of consecutive samples that crossed threshold in the velocity
            profile to detect as offset (to avoid noise-induced changes)            
        strategies: list of strategies to use
            so far, use a list containing any combination of "zero" and "velocity"
        units: str
            one of "ms", "sec", "min", "h"
        inplace: bool
            if `True`, make change in-place and return the object
            if `False`, make and return copy before making changes                                                    
        """
        fac=self._unit_fac(units)
        winsize_ms=winsize*fac
        winsize_ix=int(winsize_ms/1000.*self.fs)
        if winsize_ix % 2==0:
            winsize += 1
        min_duration_ms=min_duration*fac
        min_duration_ix=int(min_duration_ms/1000.*self.fs)        

        obj=self if inplace else self.copy()
        
        # check for unknown strategies
        for strat in strategies:
            if strat not in ["zero", "velocity"]:
                print("WARN: strategy '%s' unknown"%strat)
        
        ## detect blinks with the different strategies
        if "velocity" in strategies:
            blinks_vel=detect_blinks_velocity(self.sy, winsize_ix, vel_onset, vel_offset, min_onset_len, min_offset_len)
        else: 
            blinks_vel=np.array([])
            
        if "zero" in strategies:
            blinks_zero=detect_blinks_zero(self.sy, 1, blink_val)
        else:
            blinks_zero=np.array([])

        ## merge the two blinks
        blinks=helper_merge_blinks(blinks_vel, blinks_zero)
        obj.blinks=np.array([[on,off] for (on,off) in blinks if off-on>=min_duration_ix])
        
        obj.blink_mask=np.zeros(self.sy.size, dtype=np.int)
        
        for start,end in obj.blinks:
            obj.blink_mask[start:end]=1
        return obj    
    
    def blinks_plot(self, pdf_file: Optional[str]=None, nrow: int=5, ncol: int=3, 
                    figsize: Tuple[int,int]=(10,10), 
                    pre_blink: float=500, post_blink: float=500, units: str="ms", 
                    plot_index: bool=True):
        """
        Plot the detected blinks into separate figures each with nrow x ncol subplots. 

        Parameters
        ----------
        pdf_file: str or None
            if the name of a file is given, the figures are saved into a multi-page PDF file
        ncol: int
            number of columns for the blink-plots
        pre_blink: float
            extend plot a certain time before each blink (in ms)
        post_blink: float
            extend plot a certain time after each blink (in ms)
        units: str
            units in which the signal is plotted
        plot_index: bool
            plot a number with the blinks' index (e.g., for identifying abnormal blinks)

        Returns
        -------

        list of plt.Figure objects each with nrow*ncol subplots
        in Jupyter Notebook, those are displayed inline one after the other
        """
        fac=self._unit_fac(units)
        pre_blink_ix=int((pre_blink/1000.)*self.fs)
        post_blink_ix=int((post_blink/1000.)*self.fs)

        nblinks=self.blinks.shape[0]
        nsubplots=nrow*ncol # number of subplots per figure
        nfig=int(np.ceil(nblinks/nsubplots))

        figs=[]
        if isinstance(pdf_file,str):
            _backend=mpl.get_backend()
            mpl.use("pdf")
            plt.ioff() ## avoid showing plots when saving to PDF 
        
        iblink=0
        for i in range(nfig):
            fig=plt.figure(figsize=figsize)
            axs = fig.subplots(nrow, ncol).flatten()

            for ix,(start,end) in enumerate(self.blinks[(i*nsubplots):(i+1)*nsubplots]):
                iblink+=1
                slic=slice(start-pre_blink_ix,end+post_blink_ix)
                ax=axs[ix]
                ax.plot(self.tx[slic]*fac,self.sy[slic])

                ## highlight interpolated data
                a=np.diff(np.r_[0,self.interpolated_mask[slic],0])[:-1]
                istarts=start-pre_blink_ix+np.where(a>0)[0]
                iends=start-pre_blink_ix+np.where(a<0)[0]
                for istart,iend in zip(istarts,iends):
                    ax.axvspan(self.tx[istart]*fac,self.tx[iend]*fac,color="green", alpha=0.1)

                ## highlight blink
                ax.axvspan(self.tx[start]*fac,self.tx[end]*fac,color="red", alpha=0.2)

                if plot_index: 
                    ax.text(0.5, 0.5, '%i'%(iblink), fontsize=12, horizontalalignment='center',     
                            verticalalignment='center', transform=ax.transAxes)
            figs.append(fig)

        if pdf_file is not None:
            print("> Saving file '%s'"%pdf_file)
            with PdfPages(pdf_file) as pdf:
                for fig in figs:
                    pdf.savefig(fig)
            ## switch back to original backend and interactive mode                
            mpl.use(_backend) 
            plt.ion()
            
        return figs    

    @keephistory
    def blinks_merge(self, distance: float=100, remove_signal: bool=False, inplace=_inplace):
        """
        Merge together blinks that are close together. 
        Some subjects blink repeatedly and standard detection/interpolation can result in weird results.
        This function simply treats repeated blinks as one long blink.

        Parameters
        ----------

        distance: float
            merge together blinks that are closer together than `distance` in ms
        remove_signal: bool
            if True, set all signal values during the "new blinks" to zero so 
            that :func:`.detect_blinks()` will pick them up; interpolation will work
            either way
        inplace: bool
            if `True`, make change in-place and return the object
            if `False`, make and return copy before making changes                                                    
        """
        distance_ix=distance/self.fs*1000.

        newblinks=[] 
        i=1
        cblink=self.blinks[0,:] ## start with first blink
        while(i<self.nblinks()):
            if (self.blinks[i,0]-cblink[1])<=distance_ix:
                # merge
                cblink[1]=self.blinks[i,1]
            else:
                newblinks.append(cblink)
                cblink=self.blinks[i,:]
            i+=1            
        newblinks.append(cblink)
        newblinks=np.array(newblinks)       

        obj=self if inplace else self.copy()
        obj.blinks=newblinks

        ## set signal to zero within the new blinks
        if remove_signal:
            for start,end in obj.blinks:
                obj.sy[start:end]=0

        return obj    
    
    @keephistory
    def blinks_interpolate(self, winsize: float=11, 
                           vel_onset: float=-5, vel_offset: float=5, 
                           margin: Tuple[float,float]=(10,30), 
                           interp_type: str="cubic", inplace=_inplace):
        """
        Interpolation of missing data "in one go".
        Detection of blinks happens using Mahot (2013), see :func:`.blink_onsets_mahot()`.
        
        Parameters
        ----------
        winsize: float
            size of the Hanning-window in ms
        vel_onset: float
            velocity-threshold to detect the onset of the blink
        vel_offset: float
            velocity-threshold to detect the offset of the blink
        margin: Tuple[float,float]
            margin that is subtracted/added to onset and offset (in ms)
        interp_type: str
            type of interpolation accepted by :func:`scipy.interpolate.interp1d()`   
        inplace: bool
            if `True`, make change in-place and return the object
            if `False`, make and return copy before making changes                                                    
        """
        # parameters in sampling units (from ms)
        winsize_ix=int(np.ceil(winsize/1000.*self.fs)) 
        margin_ix=tuple(int(np.ceil(m/1000.*self.fs)) for m in margin)
        if winsize_ix % 2==0: ## ensure smoothing window is odd
            winsize_ix+=1 

        # generate smoothed signal and velocity-profile
        sym=smooth_window(self.sy, winsize_ix, "hanning")
        vel=np.r_[0,np.diff(sym)] 

        blink_onsets=blink_onsets_mahot(self.sy, self.blinks, winsize_ix, vel_onset, vel_offset,
                                        margin_ix, int(np.ceil(500/1000*self.fs)))
        obj=self if inplace else self.copy()
        obj.interpolated_mask=np.zeros(self.sy.size)
        for on,off in blink_onsets:
            obj.interpolated_mask[on:off]=1
        f=scipy.interpolate.interp1d(self.tx[obj.interpolated_mask==0], sym[obj.interpolated_mask==0], 
                                     kind=interp_type, bounds_error=False, fill_value=0)
        syr=f(self.tx)
        obj.sy=syr
        
        
        return obj
    
    @keephistory
    def blinks_interp_mahot(self, winsize: float=11, 
                           vel_onset: float=-5, vel_offset: float=5, 
                           margin: Tuple[float,float]=(10,30), 
                           blinkwindow: float=500,
                           interp_type: str="cubic",
                           plot: Optional[str]=None, 
                           plot_dim: Tuple[int,int]=(5,3),
                           plot_figsize: Tuple[int,int]=(10,8),
                           inplace=_inplace):
        """
        Implements the blink-interpolation method by Mahot (2013).
        
        Mahot, 2013:
        https://figshare.com/articles/A_simple_way_to_reconstruct_pupil_size_during_eye_blinks/688001.

        This procedure relies heavily on eye-balling (reconstructing visually convincing signal),
        so a "plot" option is provided that will plot many diagnostics (see paper linked above) that
        can help to set good parameter values for `winsize`, `vel_onset`, `vel_offset` and `margin`.

        Parameters
        ----------
        winsize: float
            size of the Hanning-window in ms
        vel_onset: float
            velocity-threshold to detect the onset of the blink
        vel_offset: float
            velocity-threshold to detect the offset of the blink
        margin: Tuple[float,float]
            margin that is subtracted/added to onset and offset (in ms)
        blinkwindow: float
            how much time before and after each blink to include (in ms)
        interp_type: str
            type of interpolation accepted by :func:`scipy.interpolate.interp1d()`
        plot: True, str or None
            if a string, the plot is going to be saved to a multipage PDF file; 
            if None, no plotting is done
            if True, plot is not saved but produced
        plot_dim: tuple nrow x ncol 
            number of subplots
        plot_figsize: tuple (width, height)
            dimensions for each figure
        inplace: bool
            if `True`, make change in-place and return the object
            if `False`, make and return copy before making changes                                                    
        """
        # parameters in sampling units (from ms)
        winsize_ix=int(np.ceil(winsize/1000.*self.fs)) 
        margin_ix=tuple(int(np.ceil(m/1000.*self.fs)) for m in margin)
        blinkwindow_ix=int(blinkwindow/1000.*self.fs)
        if winsize_ix % 2==0: ## ensure smoothing window is odd
            winsize_ix+=1 

        # generate smoothed signal and velocity-profile
        sym=smooth_window(self.sy, winsize_ix, "hanning")
        vel=np.r_[0,np.diff(sym)] 
        syr=self.sy.copy() ## reconstructed signal

        nrow,ncol=plot_dim
        nsubplots=nrow*ncol    
        nfig=int(np.ceil(self.nblinks()/nsubplots))
        figs=[]
        if isinstance(plot,str):
            _backend=mpl.get_backend()
            mpl.use("pdf")
            plt.ioff() ## avoid showing plots when saving to PDF 

        blink_onsets=blink_onsets_mahot(self.sy, self.blinks, winsize_ix, vel_onset, vel_offset,
                                           margin_ix, blinkwindow_ix)
          
        obj=self if inplace else self.copy()    
        # loop through blinks
        for ix,(onset,offset) in enumerate(blink_onsets):                
            if plot is not None:            
                if ix % nsubplots==0:
                    fig,axs=plt.subplots(nrow,ncol,figsize=plot_figsize)
                    axs=axs.flatten()
                    figs.append(fig)

            # calc the 4 time points
            t2,t3=onset,offset
            t1=max(0,t2-t3+t2)
            t4=min(t3-t2+t3, len(self)-1)
            if t1==t2:
                t2+=1
            if t3==t4:
                t3-=1
            
            txpts=[self.tx[pt] for pt in [t1,t2,t3,t4]]
            sypts=[self.sy[pt] for pt in [t1,t2,t3,t4]]
            intfct=interp1d(txpts,sypts, kind=interp_type)
            islic=slice(t2, t3)
            syr[islic]=intfct(self.tx[islic])

            ## record the interpolated datapoints
            obj.interpolated_mask[islic]=1

            slic=slice(max(0,onset-blinkwindow_ix), min(offset+blinkwindow_ix, len(self)))
            
            ## plotting for diagnostics
            #--------------------------
            if plot is not None:            
                #fig,ax1=plt.subplots()
                ax1=axs[ix % nsubplots]
                ax1.plot(self.tx[slic]/1000., self.sy[slic], color="blue", label="raw")
                ax1.plot(self.tx[slic]/1000., sym[slic], color="green", label="smoothed")
                ax1.plot(self.tx[slic]/1000., syr[slic], color="red", label="interpolated")
                ax2=ax1.twinx()
                ax2.plot(self.tx[slic]/1000., vel[slic], color="orange", label="velocity")

                for pt in (t1,t2,t3,t4):
                    ax1.plot(self.tx[pt]/1000., sym[pt], "o", color="red")
                ax1.text(0.5, 0.5, '%i'%(ix+1), fontsize=12, horizontalalignment='center',     
                    verticalalignment='center', transform=ax1.transAxes)
                if ix % nsubplots==0:
                    handles1, labels1 = ax1.get_legend_handles_labels()
                    handles2, labels2 = ax2.get_legend_handles_labels()
                    handles=handles1+handles2
                    labels=labels1+labels2
                    fig.legend(handles, labels, loc='upper right')            
        if isinstance(plot, str):
            print("> Writing PDF file '%s'"%plot)
            with PdfPages(plot) as pdf:
                for fig in figs:
                    pdf.savefig(fig)         
            ## switch back to original backend and interactive mode                
            mpl.use(_backend) 
            plt.ion()
        elif plot is not None:
            for fig in figs:
                pass
                #fig.show()

        # replace signal with the reconstructed one
        obj.sy=syr

        return obj
    
    def get_erpd(self, erpd_name: str, event_select, 
                 baseline_win: Optional[Tuple[float,float]]=None, 
                 time_win: Tuple[float,float]=(-500, 2000)):
        """
        Extract event-related pupil dilation (ERPD).
        No attempt is being made to exclude overlaps of the time-windows.

        Parameters
        ----------
        erpd_name: str
            identifier for the result (e.g., "cue-locked" or "conflict-trials")

        baseline_win: tuple (float,float) or None
            if None, no baseline-correction is applied
            if tuple, the mean value in the window in milliseconds (relative to `time_win`) is 
                subtracted from the single-trial ERPDs (baseline-correction)

        event_select: str or function
            variable describing which events to select and align to
            - if str: use all events whose label contains the string
            - if function: apply function to all labels, use those where the function returns True

        time_win: Tuple[float, float]
            time before and after event to include (in ms)


        """
        if callable(event_select):
            event_ix=np.array([bool(event_select(evlab)) for evlab in self.event_labels])
        elif isinstance(event_select, str):
            event_ix=np.array([event_select in evlab for evlab in self.event_labels])
        else:
            raise ValueError("event_select must be string or function")


        nev=event_ix.sum()
        time_win_ix=tuple(( int(np.ceil(tw/1000.*self.fs)) for tw in time_win ))
        duration_ix=time_win_ix[1]-time_win_ix[0]
        txw=np.linspace(time_win[0], time_win[1], num=duration_ix)

        ## resulting matrix and missing (interpolated/blinks/...) indicator for each datapoint
        erpd=np.zeros((nev,duration_ix))
        missing=np.ones((nev,duration_ix))

        # event-onsets as indices of the tx array
        evon=self.event_onsets[event_ix]
        # vectorized version (seems to be worse than naive one)
        #evon_ix=np.argmin(np.abs(np.tile(evon, (self.tx.size,1)).T-self.tx), axis=1)
        # naive version
        evon_ix=np.array([np.argmin(np.abs(ev-self.tx)) for ev in evon])

        for i,ev in enumerate(evon_ix):
            on,off=ev+time_win_ix[0], ev+time_win_ix[1]
            onl,offl=0,duration_ix # "local" window indices
            if on<0: ## pad with zeros in case timewindow starts before data
                onl=np.abs(on)
                on=0
            if off>=self.tx.size:
                offl=offl-(off-self.tx.size)
                off=self.tx.size

            erpd[i,onl:offl]=self.sy[on:off]
            missing[i,onl:offl]=np.logical_or(self.interpolated_mask[on:off], self.missing[on:off])

        baselines=[None for _ in range(nev)]
        if baseline_win is not None:
            if baseline_win[0]<time_win[0] or baseline_win[0]>time_win[1] or baseline_win[1]<time_win[0] or baseline_win[1]>time_win[1]:
                print("WARNING: baseline-window misspecified %s vs. %s; NOT doing baseline correction"%(baseline_win, time_win))
            else:
                blwin_ix=tuple(( np.argmin(np.abs(bw-txw)) for bw in baseline_win ))

                for i in range(nev):
                    baselines[i]=np.mean(erpd[i,blwin_ix[0]:blwin_ix[1]])
                    erpd[i,:]-=baselines[i]

        return ERPD(erpd_name, txw, erpd, missing, baselines)
    

    
#@typechecked   
class FakePupilData(PupilData):
    """
    Simulated pupil data for validation purposes.
    """
    def __init__(self,
                 pupil: PupilArray, 
                 sampling_rate: Optional[float]=None,
                 time: Optional[PupilArray]=None,
                 event_onsets: Optional[PupilArray]=None,
                 event_labels: Optional[PupilArray]=None,
                 name: Optional[str]=None,
                 sim_params: dict={},
                 real_baseline: Optional[PupilArray]=None,
                 real_response_coef: Optional[PupilArray]=None):
        """
        Constructor for artifical pupil data.
        """
        super().__init__(pupil,sampling_rate,time,event_onsets,event_labels,name)
        self.name="fake_"+self.name
        self.sim_params=sim_params
        self.sim_baseline=real_baseline
        
        ## OBS: not the real model but a simplification (npar/tmax may be different per event)
        x1=pupil_build_design_matrix(self.tx, self.event_onsets, self.fs, 
                                     sim_params["prf_npar"][0], sim_params["prf_tmax"][0], 6000)
        amp=np.mean(real_baseline)*sim_params["evoked_response_perc"]
        real_response=amp*np.dot(x1.T, real_response_coef)  ## predicted signal
        
        self.sim_response=real_response
        self.sim_response_coef=real_response_coef

    @keephistory    
    def unscale(self, mean: Optional[float]=None, sd: Optional[float]=None, inplace=_inplace):
        """
        Scale back to original values using either values provided as arguments
        or the values stored in `scale_params`.
        
        Parameters
        ----------
        mean: mean to add from signal
        sd: sd to scale with        
        inplace: bool
            if `True`, make change in-place and return the object
            if `False`, make and return copy before making changes                                                
        """
        mmean,ssd=self.scale_params["mean"],self.scale_params["sd"]
        obj=super().unscale(mean,sd,inplace)
        obj.sim_baseline=(self.sim_baseline*ssd)+mmean
        obj.sim_response=(self.sim_response*ssd)
        return obj
    
    @keephistory
    def scale(self, mean: Optional[float]=None, sd: Optional[float]=None, inplace=_inplace) -> None:
        """
        Scale the pupillary signal by subtracting `mean` and dividing by `sd`.
        If these variables are not provided, use the signal's mean and std.
        
        Parameters
        ----------
        
        mean: mean to subtract from signal
        sd: sd to scale with
        inplace: bool
            if `True`, make change in-place and return the object
            if `False`, make and return copy before making changes                                        
        
        Note
        ----
        Scaling-parameters are being saved in the `scale_params` argument. 
        """
        obj=super().scale(mean,sd)
        mean,sd=obj.scale_params["mean"],obj.scale_params["sd"]
        obj.sim_baseline=(self.sim_baseline-mean)/sd
        obj.sim_response=(self.sim_response)/sd
        return obj

    @keephistory
    def sub_slice(self, start: float=-np.inf, end: float=np.inf, units: str="sec"):
        """
        Return a new `PupilData` object that is a shortened version
        of the current one (contains all data between `start` and
        `end` in units given by `units` (one of "ms", "sec", "min", "h").
        """
        slic=super().sub_slice(start,end,units)
        evon=self.event_onsets*self._unit_fac(units)
        keepev=np.logical_and(evon>=start, evon<=end)
        slic.sim_response_coef=slic.sim_response_coef[keepev]
        return slic
            
    
    def plot(self,
             plot_range: Tuple[float,float]=(-np.infty, +np.infty),             
             interactive: bool=False, 
             baseline: bool=True, 
             response: bool=False,
             model: bool=True,
             simulated: bool=True,
             units: str="sec"
            ) -> None:
        """
        Make a plot of the pupil data using `matplotlib` or :py:func:`pypillometry.convenience.plot_pupil_ipy()`
        if `interactive=True`.

        Parameters
        ----------
        plot_range: tuple (start,end): plot from start to end (in units of `units`)
        baseline: plot baseline if estimated
        response: plot response if estimated
        model: plot full model if baseline and response have been estimated
        simulated: plot also the "ground-truth" baseline and response (i.e., the simulated one)?
        interactive: if True, plot with sliders to adjust range
        units: one of "sec"=seconds, "ms"=millisec, "min"=minutes, "h"=hours
        """
        overlays=tuple()
        overlay_labels=tuple()
        if baseline and self.baseline_estimated:
            overlays+=(self.baseline,)
            overlay_labels+=("baseline",)
        if baseline and simulated:
            overlays+=(self.sim_baseline,)
            overlay_labels+=("sim_baseline",)
        if response and self.response_estimated:
            overlays+=(self.response,)
            overlay_labels+=("response",)
        if response and simulated:
            overlays+=(self.sim_response,)
            overlay_labels+=("sim_response",)
        if model and self.baseline_estimated and self.response_estimated:
            overlays+=(self.baseline+self.response,)
            overlay_labels+=("model",)
        if model and simulated:
            overlays+=(self.sim_baseline+self.sim_response,)
            overlay_labels+=("real model",)
        self._plot(plot_range, overlays, overlay_labels, units, interactive, False, False)

        
        
def plotpd_ia(*args: PupilData, figsize: Tuple[int]=(16,8), baseline: bool=True, events: Optional[int]=0):
    """
    Interactive plotting for multiple `PupilData` objects.
    
    Parameters
    ----------
    args: `PupilData` datasets to plot
    figsize: dimensions of the plot
    baseline: plot baselines, too?
    events: plot event-markers? if None, no events are plotted, otherwise `events` 
            is the index of the `PupilData` object to take the events from
    """

    import pylab as plt
    from ipywidgets import interact, interactive, fixed, interact_manual, Layout
    import ipywidgets as widgets

    def draw_plot(plotxrange):
        xmin,xmax=plotxrange
        plt.figure(figsize=figsize)

        for i,pd in enumerate(args):
            ixmin=np.argmin(np.abs(pd.tx-xmin))
            ixmax=np.argmin(np.abs(pd.tx-xmax))


            plt.plot(pd.tx[ixmin:ixmax],pd.sy[ixmin:ixmax],label=pd.name)
            if baseline and pd.baseline_estimated:
                plt.plot(pd.tx[ixmin:ixmax], pd.baseline[ixmin:ixmax], label="BL: "+pd.name)
            
        if not events is None: 
            plt.vlines(args[events].event_onsets, *plt.ylim(), color="grey", alpha=0.5)
        plt.xlim(xmin,xmax)
        plt.legend()

    xmin=np.min([pd.tx.min() for pd in args])
    xmax=np.max([pd.tx.max() for pd in args])
    wid_range=widgets.FloatRangeSlider(
        value=[xmin,xmax],
        min=xmin,
        max=xmax,
        step=1,
        description=' ',
        disabled=False,
        continuous_update=False,
        orientation='horizontal',
        readout=True,
        readout_format='.1f',
        layout=Layout(width='100%', height='80px')
    )

    interact(draw_plot, plotxrange=wid_range)
    
    

    
def plotpd(*args: PupilData, subplots: bool=False, baseline: bool=False):
    """
    Plotting for `PupilData` objects.
    
    Parameters
    ----------
    
    subplots: plot the different `PupilData`-objects in the same plot or subplots
    """
    if len(args)<3:
        ncol=len(args)
        nrow=1
    else:
        ncol=3
        nrow=np.ceil(len(args)/3.0)
    for i,pd in enumerate(args):
        if subplots:
            plt.subplot(nrow,ncol,i+1)
            plt.title(pd.name)
        plt.plot(pd.tx/1000./60., pd.sy, label=pd.name)
        if baseline and pd.baseline_estimated:
            plt.plot(pd.tx/1000./60., pd.baseline, label="BL: "+pd.name)
        if i==0:
            plt.xlabel("time (min)")
            plt.ylabel("PD")
    if not subplots:
        plt.legend()
        
    
def create_fake_pupildata(**kwargs):
    """
    Return a :py:class:`pyillometry.pupildata.FakePupilData` object by buildling it with
    :py:func:`pypillometry.fakedata.get_dataset()`.
    
    Parameters
    -----------
    
    ntrials:int
        number of trials
    isi: float
        inter-stimulus interval in seconds
    rtdist: tuple (float,float)
        mean and std of a (truncated at zero) normal distribution to generate response times
    pad: float
        padding before the first and after the last event in seconds        
    fs: float
        sampling rate in Hz
    baseline_lowpass: float
        cutoff for the lowpass-filter that defines the baseline
        (highest allowed frequency in the baseline fluctuations)        
    evoked_response_perc: float
        amplitude of the pupil-response as proportion of the baseline     
    response_fluct_sd: float
        How much do the amplitudes of the individual events fluctuate?
        This is determined by drawing each individual pupil-response to 
        a single event from a (positive) normal distribution with mean as determined
        by `evoked_response_perc` and sd `response_fluct_sd` (in units of 
        `evoked_response_perc`).
    prf_npar: tuple (float,float)
        (mean,std) of the npar parameter from :py:func:`pypillometry.pupil.pupil_kernel()`. 
        If the std is exactly zero, then the mean is used for all pupil-responses.
        If the std is positive, npar is taken i.i.d. from ~ normal(mean,std) for each event.
    prf_tmax: tuple (float,float)
        (mean,std) of the tmax parameter from :py:func:`pypillometry.pupil.pupil_kernel()`. 
        If the std is exactly zero, then the mean is used for all pupil-responses.
        If the std is positive, tmax is taken i.i.d. from ~ normal(mean,std) for each event.
    prop_spurious_events: float
        Add random events to the pupil signal. `prop_spurious_events` is expressed
        as proportion of the number of real events. 
    noise_amp: float
        Amplitude of random gaussian noise that sits on top of the simulated signal.
        Expressed in units of mean baseline pupil diameter.
    """
    sim_params={
        "ntrials":100,
        "isi":1000.0,
        "rtdist":(1000.0,500.0),
        "pad":5000.0,
        "fs":1000.0,
        "baseline_lowpass":0.1,
        "evoked_response_perc":0.001,
        "response_fluct_sd":1,
        "prf_npar":(10.35,0),
        "prf_tmax":(917.0,0),
        "prop_spurious_events":0.1,
        "noise_amp":0.0001
    }
    sim_params.update(kwargs)
    #print(sim_params)
    tx,sy,baseline,event_onsets,response_coef=get_dataset(**sim_params)
    event_labels=["event" for _ in range(event_onsets.size)]
    ds=FakePupilData(sy,sim_params["fs"],tx, event_onsets,event_labels=event_labels,
                     sim_params=sim_params, 
                     real_baseline=baseline, real_response_coef=response_coef)
    return ds
    
    


"""
io.py
=====

Read/Write data from/to disk.
"""

try:
   import cPickle as pickle
except:
   import pickle
import requests
from .pupildata import *

def pd_write_pickle(pdobj, fname):
    """
    Store the :class:`.PupilData`-object `pdobj` in file using :mod:`pickle`.
    
    Parameters
    ----------
    
    pdobj: :class:`.PupilData`
        dataset to save
    fname: str
        filename to save to
    """
    with open(fname, "wb") as f:
        pickle.dump(pdobj,f)
    
def pd_read_pickle(fname):
    """
    Read the :class:`.PupilData`-object `pdobj` from file using :mod:`pickle`.
    
    Parameters
    ----------
    
    fname: str
        filename or URL to load data from
        
    Returns
    -------
    
    pdobj: :class:`.PupilData`
        loaded dataset 
    """
    if fname.startswith("http"):
        # try loading from URL
        res=requests.get(fname)
        if res.status_code==200:
            pdobj=pickle.loads(res.content)
    else:
        with open(fname, 'rb') as f:
            pdobj=pickle.load(f)
    return pdobj



"""
erpd.py
============

Event-related pupil dilation.

"""
from .io import *

import pylab as plt
import matplotlib as mpl
from matplotlib.backends.backend_pdf import PdfPages
import numpy as np
import scipy.signal as signal
from scipy.interpolate import interp1d
from scipy import interpolate
import scipy


from typing import Sequence, Union, List, TypeVar, Optional, Tuple, Callable
PupilArray=Union[np.ndarray, List[float]]
import collections.abc


class ERPD:
    """
    Class representing a event-related pupillary dilation (ERPD) for one subject.
    """
    def __init__(self, name, tx, erpd, missing, baselines):
        self.name=name
        self.baselines=baselines
        self.tx=tx
        self.erpd=erpd
        self.missing=missing

    def summary(self) -> dict:
        """Return a summary of the :class:`.PupilData`-object."""
        summary=dict(
            name=self.name,
            nevents=self.erpd.shape[0],
            n=self.erpd.shape[1],
            window=(self.tx.min(), self.tx.max())
        )
        return summary

    def write_file(self, fname:str):
        """
        Save to file (using :mod:`pickle`).
        
        Parameters
        ----------
        
        fname: str
            filename
        """
        pd_write_pickle(self, fname)
       
    @classmethod
    def from_file(cls, fname:str):
        """
        Reads a :class:`.ERPD` object from a pickle-file.
        Use as ``pypillometry.ERPD.from_file("yourfile.pd")``.
        
        Parameters
        ----------
        
        fname: str
            filename
        """
        r=pd_read_pickle(fname)
        return r
    
    def __repr__(self) -> str:
        """Return a string-representation of the dataset."""
        pars=self.summary()
        del pars["name"]
        s="ERPD({name}):\n".format(name=self.name)
        flen=max([len(k) for k in pars.keys()])
        for k,v in pars.items():
            s+=(" {k:<"+str(flen)+"}: {v}\n").format(k=k,v=v)
        return s
                
    def plot(self, overlays=None, meanfct=np.mean, varfct=scipy.stats.sem, plot_missing: bool=True): 
        """
        Plot mean and error-ribbons using `varct`.
        
        Parameters
        ----------
        
        overlays: single or sequence of :class:`.ERPDSingleSubject`-objects 
            the overlays will be added to the same plot
        
        meanfct: callable
            mean-function to apply to the single-trial ERPDs for plotting
        varfct: callable or None
            function to calculate error-bands (e.g., :func:`numpy.std` for standard-deviation 
            or :func:`scipy.stats.sem` for standard-error)
            if None, no error bands are plotted
            
        plot_missing: bool
            plot percentage interpolated/missing data per time-point?
        """
        merpd=meanfct(self.erpd, axis=0)
        sderpd=varfct(self.erpd, axis=0) if callable(varfct) else None
        percmiss=np.mean(self.missing, axis=0)*100.
        ax1=plt.gca()        
        if sderpd is not None:
            ax1.fill_between(self.tx, merpd-sderpd, merpd+sderpd, color="grey", alpha=0.3)
        ax1.plot(self.tx, merpd, label=self.name)        
        ax1.axvline(x=0, color="red")        
        ax1.set_ylabel("mean PD")
        ax1.set_xlabel("time (ms)")
        ax1.set_title(self.name)
        if plot_missing:
            ax2=ax1.twinx()
            ax2.plot(self.tx, percmiss, alpha=0.3)
            ax2.set_ylim(0,100)
            ax2.set_ylabel("% missing")
            
        if overlays is not None:
            if not isinstance(overlays, collections.abc.Sequence):
                overlays=[overlays]
            for ov in overlays:
                merpd=meanfct(ov.erpd, axis=0)
                sderpd=varfct(ov.erpd, axis=0) if callable(varfct) else None
                percmiss=np.mean(ov.missing, axis=0)*100.
                if sderpd is not None:
                    ax1.fill_between(self.tx, merpd-sderpd, merpd+sderpd, color="grey", alpha=0.3)
                ax1.plot(self.tx, merpd, label=ov.name)        
                if plot_missing:
                    ax2.plot(ov.tx, percmiss, alpha=0.3)
        ax1.legend()
        

def plot_erpds(erpds):
    """
    Plot a list of ERPD objects.
    """
    erpds[0].plot(erpds[1:len(erpds)])
    

def group_erpd(datasets: List, erpd_name: str, event_select, 
               baseline_win: Optional[Tuple[float,float]]=None, 
               time_win: Tuple[float,float]=(-500, 2000),
               subj_meanfct=np.mean):
    """
    Calculate group-level ERPDs by applying `subj_meanfct` to each subj-level ERPD.
    

    Parameters
    ----------
    datasets: list of PupilData objects
        one PupilData object for each subject that should go into the group-level ERPD.
        
    erpd_name: str
        identifier for the result (e.g., "cue-locked" or "conflict-trials")

    baseline_win: tuple (float,float) or None
        if None, no baseline-correction is applied
        if tuple, the mean value in the window in milliseconds (relative to `time_win`) is 
            subtracted from the single-trial ERPDs (baseline-correction)

    event_select: str or function
        variable describing which events to select and align to
        - if str: use all events whose label contains the string
        - if function: apply function to all labels, use those where the function returns True

    time_win: Tuple[float, float]
        time before and after event to include (in ms)
        
    subj_meanfct: fct
        function to summarize each individual ERPD
        
    Returns
    -------

    an :py:class:`ERPD` object for the group
    """        
    erpds=[d.get_erpd(erpd_name, event_select, baseline_win, time_win) for d in datasets]
    merpd=np.array([subj_meanfct(e.erpd, axis=0) for e in erpds])
    mmissing=np.array([subj_meanfct(e.missing, axis=0) for e in erpds])
    tx=erpds[0].tx
    erpd=ERPD(erpd_name, tx, merpd, mmissing, None)    
    return erpd



"""
pupil.py
========

functions related to pupillary responses.
"""
import numpy as np
import scipy.optimize
import pylab as plt
from .convenience import *

def pupil_kernel_t(t,npar,tmax):
    """
    According to Hoeks and Levelt (1993, 
    https://link.springer.com/content/pdf/10.3758%2FBF03204445.pdf), 
    the PRF can be described by the so-called Erlang gamma function
    $$h_{HL}(t) = t^n e^{-nt/t_{max}}$$
    which we normalize to
    $$h(t)=\\frac{1}{h_{max}} h_{HL}(t)$$
    where $$h_{max} = \max_t{\\left(h_{HL}(t)\\right)} = e^{-n}t_{max}^{n}$$
    which yields a maximum value of 1 for this function. 
    The function $h(t)$ is implemented in :py:func:`pupil_kernel()`.
    
    This version of the function evaluates the PRF at inputs `t`.
    
    Parameters
    -----------
    t: float/np.array
        in ms
    npar: float
        n in the equation above
    tmax: float
        t_{max} in the equation above

    Returns
    --------
    h: np.array
        PRF evaluated at `t`
    """
    t=np.array(t)
    npar=float(npar)
    tmax=float(tmax)
    hmax=np.exp(-npar)*tmax**npar ## theoretical maximum
    h = t**(npar) * np.exp(-npar*t / tmax)   #Erlang gamma function Hoek & Levelt (1993)
    h=h/hmax
    return h

def pupil_get_max_duration(npar,tmax,thr=1e-8,stepsize=1.):
    """
    Get the time when the PRF with parameters $n$ and $t_{max}$ is decayed to
    `thr`. This gives an indication of how long the `duration` parameter
    in :py:func:`pupil_kernel()` should be chosen.
    
    Parameters
    -----------
    npar,tmax: float
        PRF parameters, see :py:func:`pupil_kernel()`
    thr: float
        desired value to which the PRF is decayed (the lower `thr`, the longer the duration)
    stepsize: float
        precision of the maximum duration in ms
        
    Returns
    --------
    tdur: float
        first value of t so that PRF(t)<`thr`
    """
    # start looking from `tmax` (which is time of the peak)
    tdur=tmax
    while pupil_kernel_t(tdur,npar,tmax)>thr:
        tdur=tdur+stepsize # in steps of `stepsize` ms
    return tdur

def pupil_kernel(duration=4000, fs=1000, npar=10.1, tmax=930.0):
    """
    According to Hoeks and Levelt (1993, 
    https://link.springer.com/content/pdf/10.3758%2FBF03204445.pdf), 
    the PRF can be described by the so-called Erlang gamma function
    $$h_{HL}(t) = t^n e^{-nt/t_{max}}$$
    which we normalize to
    $$h(t)=\\frac{1}{h_{max}} h_{HL}(t)$$
    where $$h_{max} = \max_t{\\left(h_{HL}(t)\\right)} = e^{-n}t_{max}^{n}$$
    which yields a maximum value of 1 for this function. 
    The function $h(t)$ is implemented in `pp.pupil_kernel()`.
    
    Parameters
    -----------
    
    duration: float
        in ms; maximum of the time window for which to calculate the PRF [0,duration]
    fs: float
        sampling rate for resolving the PRF
    npar: float
        n in the equation above
    tmax: float
        t_{max} in the equation above
        
    Returns
    --------
    
    h: np.array
        sampled version of h(t) over the interval [0,`duration`] with sampling rate `fs`
    """
    n=int(duration/1000.*fs)
    t = np.linspace(0,duration, n, dtype = np.float)
    h=pupil_kernel_t(t,npar,tmax)
    #h = t**(npar) * np.exp(-npar*t / tmax)   #Erlang gamma function Hoek & Levelt (1993)
    #hmax=np.exp(-npar)*tmax**npar ## theoretical maximum
    return h#/h.max() # rescale to height=1

def plot_prf(npar=10.1,tmax=930,max_duration="estimate",fs=500,**kwargs):
    """
    Plot profile of the pupil-response function (PRF) with 
    parameters `npar` and `tmax`.
    """
    if max_duration=="estimate":
        max_duration=pupil_get_max_duration(npar,tmax)
    n=int(fs*(max_duration/1000.))
    tx=np.linspace(0,max_duration,n)
    prf=pupil_kernel_t(tx,npar,tmax)
    plt.plot(tx,prf,**kwargs)
    plt.xlabel("time [ms]")
    plt.ylabel("AU")
    

def pupil_build_design_matrix(tx,event_onsets,fs,npar,tmax,max_duration="estimate"):
    """
    Construct design matrix (nevents x n).
    Each column has a single pupil-kernel with parameters `npar`, `tmax` starting at 
    each `event_onset`.
    
    Parameters
    ----------
    
    tx: np.array
        in ms
    event_onsets: np.array
        timing of the events
    fs: float
        sampling rate (Hz)
    npar: float
        n in the equation of :py:func:`pypillometry.pupil.pupil_kernel()`
    tmax: float
        t_{max} in :py:func:`pypillometry.pupil.pupil_kernel()`
    max_duration: float or "estimate"
        either maximum duration in milliseconds or string "estimate" which
        causes the function to determine an appropriate duration based on the
        kernel parameters `npar` and `tmax`
        
    Returns
    -------
    x1: np.array (nevents x n) 
        design matrix
    """
    if max_duration=="estimate":
        max_duration=pupil_get_max_duration(npar,tmax)

        
    h=pupil_kernel(duration=max_duration, fs=fs, npar=npar, tmax=tmax) ## pupil kernel

    # event-onsets for each event
    x1 = np.zeros((event_onsets.size, tx.size), dtype=np.float) # onsets

    # event-onsets as indices of the txd array
    evon_ix=np.argmin(np.abs(np.tile(event_onsets, (tx.size,1)).T-tx), axis=1)

    for i in range(evon_ix.size):
        slic_add=h.size if (evon_ix[i]+h.size)<x1.shape[1] else x1.shape[1]-evon_ix[i]            
        x1[i,evon_ix[i]:evon_ix[i]+slic_add]=h[0:slic_add]
                
      
    ## old, vectorized version (I thought it would be faster but it is, in fact, a lot slower :-(
    # # prepare stimulus and response-regressors
    # h=pupil_kernel(duration=max_duration, fs=fs, npar=npar, tmax=tmax) ## pupil kernel
    # 
    # # event-onsets for each event
    # x1 = np.zeros((event_onsets.size, tx.size), dtype=np.float) # onsets
    # 
    # # event-onsets as indices of the txd array
    # evon_ix=np.argmin(np.abs(np.tile(event_onsets, (tx.size,1)).T-tx), axis=1)
    # 
    # X=np.meshgrid(np.arange(x1.shape[1]), np.arange(x1.shape[0]))[0]
    # evon_ix_M1=np.tile(evon_ix, (x1.shape[1],1)).T
    # evon_ix_M2=np.tile(evon_ix+h.size, (x1.shape[1],1)).T
    # 
    # x1[ np.arange(event_onsets.size), evon_ix ]=1
    # x1[ np.logical_and(X>=evon_ix_M1, X<evon_ix_M2) ]=np.tile(h, evon_ix.size)
    return x1

def pupil_response(tx, sy, event_onsets, fs, npar="free", tmax="free", verbose=10, 
                   bounds={"npar":(1,20), "tmax":(100,2000)}, display_progress=True):
    """
    Estimate pupil-response based on event-onsets.
    
    tx : np.ndarray
        time-vector in milliseconds        
    sy : np.ndarray
        (baseline-corrected) pupil signal
    event_onsets : list
        onsets of events (stimuli/responses) in milliseconds        
    fs : float
        sampling rate in Hz
    npar: float
        npar-parameter for the canonical response-function or "free";
        in case of "free", the function optimizes for this parameter
    tmax: float
        tmax-parameter for the canonical response-function or "free";
        in case of "free", the function optimizes for this parameter
    bounds: dict
        in case that one or both parameters are estimated, give the lower
        and upper bounds for the parameters
    """
    def vprint(v, s):
        if v<=verbose:
            print(s,end="")
     
    if npar=="free" and tmax=="free":
        print("MSG: optimizing both npar and tmax, might take a while...")
        def objective(x, event_onsets, tx,sy,fs):
            vprint(50,".")
            npar_t,tmax_t=x
            npar,tmax=x
            #npar=trans_logistic_vec(npar_t, a=bounds["npar"][0], b=bounds["npar"][1], inverse=True)
            #tmax=trans_logistic_vec(tmax_t, a=bounds["tmax"][0], b=bounds["tmax"][1], inverse=True)
            maxdur=pupil_get_max_duration(npar,tmax)
            vprint(100, "\nnpar,tmax,maxdur=(%.2f,%.2f,%i)"%(npar,tmax,maxdur))            
            x1=pupil_build_design_matrix(tx, event_onsets, fs, npar, tmax, maxdur)
            coef=scipy.optimize.nnls(x1.T, sy)[0]    
            pred=np.dot(x1.T, coef)  ## predicted signal
            resid=sy-pred         ## residuals            
            return np.sum(resid**2)
        
        #npar_start_trans=trans_logistic_vec(10,a=bounds["npar"][0], b=bounds["npar"][1],inverse=False)
        #tmax_start_trans=trans_logistic_vec(900,a=bounds["tmax"][0], b=bounds["tmax"][1],inverse=False)
        #r=scipy.optimize.minimize(objective, (npar_start_trans, tmax_start_trans), 
        #                          args=(event_onsets,tx,sy,fs), 
        #                          method="Nelder-Mead")    
        r=scipy.optimize.minimize(objective, (10,900), #(npar_start_trans, tmax_start_trans), 
                                  args=(event_onsets,tx,sy,fs), bounds=[bounds["npar"],bounds["tmax"]],
                                 options={"disp":display_progress})
                                  #method="Nelder-Mead")    
        
        #npar=trans_logistic_vec(r.x[0], a=bounds["npar"][0], b=bounds["npar"][1], inverse=False)
        #tmax=trans_logistic_vec(r.x[1], a=bounds["tmax"][0], b=bounds["tmax"][1], inverse=False)
        npar,tmax=r.x[0],r.x[1]
    elif npar=="free":
        print("MSG: optimizing npar only, might take a while...")
        def objective(x, tmax, event_onsets, tx,sy,fs):
            vprint(50,".")
            npar=x
            maxdur=pupil_get_max_duration(npar,tmax)
            vprint(100, "\nnpar,maxdur=(%.2f,%i)"%(npar,maxdur))            
            x1=pupil_build_design_matrix(tx, event_onsets, fs, npar, tmax, maxdur)            
            coef=scipy.optimize.nnls(x1.T, sy)[0]    
            pred=np.dot(x1.T, coef)  ## predicted signal
            resid=sy-pred         ## residuals            
            return np.sum(resid**2)
        r=scipy.optimize.minimize_scalar(objective, bounds=bounds["npar"],
                                  args=(tmax,event_onsets,tx,sy,fs), 
                                  method="bounded", options={"disp":display_progress,"xatol":.1})        
        npar=r.x
    elif tmax=="free":
        print("MSG: optimizing tmax only, might take a while...")
        def objective(x, npar, event_onsets, tx,sy,fs):
            vprint(50,".")            
            tmax=x
            maxdur=pupil_get_max_duration(npar,tmax)    
            vprint(100, "\ntmax,maxdur=(%.2f,%i)"%(tmax,maxdur))            
            x1=pupil_build_design_matrix(tx, event_onsets, fs, npar, tmax, maxdur)
            coef=scipy.optimize.nnls(x1.T, sy)[0]    
            pred=np.dot(x1.T, coef)  ## predicted signal
            resid=sy-pred         ## residuals            
            return np.sum(resid**2)
        r=scipy.optimize.minimize_scalar(objective, bounds=bounds["tmax"],
                                  args=(npar,event_onsets,tx,sy,fs), 
                                  method="bounded",options={"disp":display_progress,"xatol":1})        
        tmax=r.x
  
    maxdur=pupil_get_max_duration(npar,tmax)
    x1=pupil_build_design_matrix(tx, event_onsets, fs, npar, tmax, maxdur)
    coef=scipy.optimize.nnls(x1.T, sy)[0]    
    pred=np.dot(x1.T, coef)  ## predicted signal
    
    return pred, coef, npar, tmax, x1

        
     
def pupilresponse_nnls(tx, sy, event_onsets, fs, npar=10.1, tmax=930):
    """
    Estimate single-event pupil responses based on canonical PRF (`pupil_kernel()`)
    using non-negative least-squares (NNLS).
        
    Parameters
    -----------
    
    tx : np.ndarray
        time-vector in milliseconds
        
    sy : np.ndarray
        (baseline-corrected) pupil signal
        
    event_onsets : list
        onsets of events (stimuli/responses) in seconds
        
    fs : float
        sampling rate in Hz
        
    npar,tmax: float
        parameters for :py:func:`pypillometry.pupil.pupil_kernel()`
        
    Returns
    --------
    
    (coef,pred,resid): tuple
        coef: purely-positive regression coefficients
        pred: predicted signal
        resid: residuals (sy-pred)
    """
    x1=pupil_build_design_matrix(tx, event_onsets, fs, npar, tmax, "estimate")
    
    ## we use a non-negative least squares solver to force the PRF-coefficients to be positive
    coef=scipy.optimize.nnls(x1.T, sy)[0]    
    pred=np.dot(x1.T, coef)  ## predicted signal
    resid=sy-pred         ## residual

    return coef,pred,resid
    

def stat_event_interval(tx,sy,event_onsets,interval,statfct=np.mean):
    """
    Return result of applying a statistical function to pupillometric data in a
    given interval relative to event-onsets. For example, extract mean 
    pupil-size in interval before trial onset.
        
    Parameters
    -----------
    
    tx : np.ndarray
        time-vector in milliseconds
        
    sy : np.ndarray
        (baseline-corrected) pupil signal
        
    event_onsets : list
        onsets of events (stimuli/responses) in seconds
        
    interval : tuple (min,max)
        time-window in ms relative to event-onset (0 is event-onset)
    
    statfct : function
        function mapping np.array to a single number
    
    Returns
    --------
    
    result: np.array
        number of event-onsets long result array
    """
    event_onsets=np.array(event_onsets)
    starts=event_onsets+interval[0]
    ends  =event_onsets+interval[1]

    res=np.zeros(event_onsets.size)

    for i,interv in enumerate(zip(starts,ends)):
        start_ix=np.argmin(np.abs(interv[0]-tx))
        end_ix=np.argmin(np.abs(interv[1]-tx))
        if start_ix==end_ix:
            end_ix+=1
        res[i]=statfct(sy[start_ix:end_ix])
    return res
    
        
    


"""
Pypillometry
============

This is a python-package to help with processing of pupillometric data.
"""
from .baseline import *
from .convenience import *
from .fakedata import *
from .pupil import *
from .pupildata import *
from .erpd import *

import os.path
__package_path__ = os.path.abspath(os.path.dirname(__file__))




"""
fakedata.py
====================================
Generate artificial pupil-data.
"""
import numpy as np
import scipy.stats as stats

from .baseline import *
from .pupil import *

def generate_pupil_data(event_onsets, fs=1000, pad=5000, baseline_lowpass=0.2, 
                        evoked_response_perc=0.02, response_fluct_sd=1,
                        prf_npar=(10.35,0), prf_tmax=(917.0,0),
                        prop_spurious_events=0.2, noise_amp=0.0005):
    """
    Generate artificial pupil data as a sum of slow baseline-fluctuations
    on which event-evoked responses are "riding". 
    
    Parameters
    -----------
    
    event_onsets: list
        list of all events that evoke a response (in seconds)
        
    fs: float
        sampling rate in Hz
    pad: float
        append `pad` milliseconds of signal after the last event is decayed    
    baseline_lowpass: float
        cutoff for the lowpass-filter that defines the baseline
        (highest allowed frequency in the baseline fluctuations)
        
    evoked_response_perc: float
        amplitude of the pupil-response as proportion of the baseline 
    
    response_fluct_sd: float
        How much do the amplitudes of the individual events fluctuate?
        This is determined by drawing each individual pupil-response to 
        a single event from a (positive) normal distribution with mean as determined
        by `evoked_response_perc` and sd `response_fluct_sd` (in units of 
        `evoked_response_perc`).
    prf_npar: tuple (float,float)
        (mean,std) of the npar parameter from :py:func:`pypillometry.pupil.pupil_kernel()`. 
        If the std is exactly zero, then the mean is used for all pupil-responses.
        If the std is positive, npar is taken i.i.d. from ~ normal(mean,std) for each event.
    prf_tmax: tuple (float,float)
        (mean,std) of the tmax parameter from :py:func:`pypillometry.pupil.pupil_kernel()`. 
        If the std is exactly zero, then the mean is used for all pupil-responses.
        If the std is positive, tmax is taken i.i.d. from ~ normal(mean,std) for each event.
    prop_spurious_events: float
        Add random events to the pupil signal. `prop_spurious_events` is expressed
        as proportion of the number of real events. 
        
    noise_amp: float
        Amplitude of random gaussian noise that sits on top of the simulated signal.
        Expressed in units of mean baseline pupil diameter.
        
    
    Returns
    --------
    
    tx, sy: np.array
        time and simulated pupil-dilation (n)
    x0: np.array
        baseline (n)
    delta_weights: np.array
        pupil-response strengths (len(event_onsets))
    """
    nevents=len(event_onsets)
    ## npar
    if prf_npar[1]==0: # deterministic parameter
        npars=np.ones(nevents)*prf_npar[0]
    else:
        npars=np.random.randn(nevents)*prf_npar[1]+prf_npar[0]

    ## tmax
    if prf_tmax[1]==0: # deterministic parameter
        tmaxs=np.ones(nevents)*prf_tmax[0]
    else:
        tmaxs=np.random.randn(nevents)*prf_tmax[1]+prf_tmax[0]

    if np.any(npars<=0):
        raise ValueError("npar must be >0")
    if np.any(tmaxs<=0):
        raise ValueError("tmax must be >0")

    # get maximum duration of one of the PRFs
    maxdur=pupil_get_max_duration(npars.min(), tmaxs.max())

    T=np.array(event_onsets).max()+maxdur+pad # stop pad millisec after last event
    n=int(np.ceil(T/1000.*fs)) # number of sampling points
    sy=np.zeros(n)       # pupil diameter 
    tx=np.linspace(0,T,n)     # time-vector in milliseconds

    # create baseline-signal 
    slack=int(0.50*n) # add slack to avoid edge effects of the filter
    x0=butter_lowpass_filter(np.random.rand(n+slack), baseline_lowpass, fs, 2)[slack:(n+slack)]
    x0=x0*1000+5000 # scale it up to a scale as usually obtained from eyetracker


    ### real events regressor
    ## scaling
    event_ix=(np.array(event_onsets)/1000.*fs).astype(np.int)
    #a, b = (myclip_a - my_mean) / my_std, (myclip_b - my_mean) / my_std
    delta_weights=stats.truncnorm.rvs(-1/response_fluct_sd,np.inf, loc=1, scale=response_fluct_sd, size=event_ix.size)
    x1=np.zeros_like(sy)

    for i,ev in enumerate(event_onsets):
        # create kernel and delta-functions for events
        kernel=pupil_kernel(duration=maxdur,fs=fs,npar=npars[i], tmax=tmaxs[i])
        x1[event_ix[i]:(event_ix[i]+kernel.size)]=x1[event_ix[i]:(event_ix[i]+kernel.size)]+kernel*delta_weights[i]

    ## spurious events regressor

    sp_event_ix=np.random.randint(low=0,high=np.ceil((T-maxdur-pad)/1000.*fs),size=int( nevents*prop_spurious_events ))
    sp_events=tx[ sp_event_ix ]
    n_sp_events=sp_events.size

    ## npar
    if prf_npar[1]==0: # deterministic parameter
        npars=np.ones(n_sp_events)*prf_npar[0]
    else:
        npars=np.random.randn(n_sp_events)*prf_npar[1]+prf_npar[0]

    ## tmax
    if prf_tmax[1]==0: # deterministic parameter
        tmaxs=np.ones(n_sp_events)*prf_tmax[0]
    else:
        tmaxs=np.random.randn(n_sp_events)*prf_tmax[1]+prf_tmax[0]


    ## scaling
    sp_delta_weights=stats.truncnorm.rvs(-1/response_fluct_sd,np.inf, loc=1, scale=response_fluct_sd, size=sp_event_ix.size)
    x2=np.zeros_like(sy)

    for i,ev in enumerate(sp_events):
        # create kernel and delta-functions for events
        kernel=pupil_kernel(duration=maxdur,fs=fs,npar=npars[i], tmax=tmaxs[i])
        x2[sp_event_ix[i]:(sp_event_ix[i]+kernel.size)]=x2[sp_event_ix[i]:(sp_event_ix[i]+kernel.size)]+kernel*sp_delta_weights[i]

    amp=np.mean(x0)*evoked_response_perc # mean amplitude for the evoked response
    noise=noise_amp*np.mean(x0)*np.random.randn(n)

    sy = x0 + amp*x1 + amp*x2 + noise

    return (tx,sy,x0,delta_weights)


def get_dataset(ntrials=100, isi=2000, rtdist=(1000,500),fs=1000,pad=5000, **kwargs):
    """
    Convenience function to run :py:func:`generate_pupil_data()` with standard parameters.
    Parameters
    -----------
    
    ntrials:int
        number of trials
    isi: float
        inter-stimulus interval in milliseconds
    rtdist: tuple (float,float)
        mean and std of a (truncated at zero) normal distribution to generate response times
    fs: float
        sampling rate
    pad: float
        padding before the first and after the last event in seconds
    kwargs: dict
        arguments for :py:func:`pypillometry.fakedata.generate_pupil_data()`

    
    Returns
    --------
        
    tx, sy: np.array
        time and simulated pupil-dilation (n)
    baseline: np.array
        baseline (n)
    event_onsets: np.array
        timing of the simulated event-onsets (stimuli and responses not separated)
    response_coef: np.array
        pupil-response strengths (len(event_onsets))
    """
    stim_onsets=np.arange(ntrials)*isi+pad
    rts=stats.truncnorm.rvs( (0-rtdist[0])/rtdist[1], np.inf, loc=rtdist[0], scale=rtdist[1], size=ntrials)
    resp_onsets=stim_onsets+rts
    event_onsets=np.concatenate( (stim_onsets, resp_onsets) )

    kwargs.update({"fs":fs})
    tx,sy,baseline,response_coef=generate_pupil_data(event_onsets, **kwargs)
    return tx,sy,baseline,event_onsets, response_coef



import unittest
import tempfile
import os, pickle, hashlib
import sys
#sys.path.insert(0,"..")
#import pypillometry as pp
from .. import *

class TestIO(unittest.TestCase):
    def test_pd_read_pickle_file(self):
        d=pd_read_pickle("data/test.pd")
        self.assertEqual(d.fs, 500.0)
        self.assertEqual(len(d), 60001)
    def test_pd_read_pickle_http(self):
        d=pd_read_pickle("https://github.com/ihrke/pypillometry/blob/master/data/test.pd?raw=true")
        self.assertEqual(d.fs, 500.0)
        self.assertEqual(len(d), 60001)
    
    def test_pd_write_pickle(self):
        d=pd_read_pickle("data/test.pd")#create_fake_pupildata(ntrials=10)
        fpath=tempfile.mkdtemp()
        fname=os.path.join(fpath, "test2.pd")
        pd_write_pickle(d, fname)
        x=pd_read_pickle(fname)
        self.assertEqual(x.size_bytes(), d.size_bytes())
        self.assertEqual(x.name, d.name)
        dmd5=hashlib.md5(pickle.dumps(d,-1)).hexdigest()
        xmd5=hashlib.md5(pickle.dumps(x,-1)).hexdigest()
        self.assertEqual(dmd5,xmd5)

if __name__ == '__main__':
    unittest.main()


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("requirements.txt") as fh:
    requirements=fh.read().split()

with open("VERSION") as fh:
    version=fh.read().strip()
    
setuptools.setup(
    name="pypillometry",
    version=version,
    author="Matthias Mittner",
    author_email="matthias.mittner@uit.no",
    description="Pupillometry in Python.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ihrke/pypillometry",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=requirements,
    python_requires='>=3.6',
)



import unittest
import sys
#sys.path.insert(0,"..")
#import pypillometry as pp
from .. import *

class TestPupil(unittest.TestCase):
    def setUp(self):
        self.d=create_fake_pupildata(ntrials=100)
    def test_pupil_kernel_t(self):
        pupil_kernel_t([1,2], 10, 900)

if __name__ == '__main__':
    unittest.main()


"""
baseline.py
===========

Functions for estimating tonic pupillary fluctuations (baseline).
"""
import numpy as np
import scipy.signal as signal
import scipy
import math

import scipy.interpolate
from scipy.interpolate import interp1d, splrep, splev

from .pupil import *
from .convenience import *

stan_code_baseline_model_asym_laplac="""
// Stan model for pupil-baseline estimation
//
functions{
    // asymmetric laplace function with the mu, sigma, tau parametrization
    real skew_double_exponential_lpdf(real y, real mu, real sigma, real tau) {
      return log(tau) + log1m(tau)
        - log(sigma)
        - 2 * ((y < mu) ? (1 - tau) * (mu - y) : tau * (y - mu)) / sigma;
    }
    
    // zero-centered asymmetric laplace function with the mu, lambda, kappa parametrization
    real skew_double_exponential2_lpdf(real y, real lam, real kappa) {
      return log(lam) - log(kappa+1/kappa)
        + ((y<0) ? (lam/kappa) : (-lam*kappa))*(y);
    }
}
data{ 
  int<lower=1> n; // number of timepoints in the signal
  vector[n] sy;   // the pupil signal
    
  int<lower=1> ncol; // number of basis functions (columns in B)
  matrix[n,ncol] B;  // spline basis functions
    
  int<lower=1> npeaks;          // number of lower peaks in the signal
  int<lower=1> peakix[npeaks];  // index of the lower peaks in sy
  vector<lower=0>[npeaks] lam_prominences; // lambda-converted prominence values
    
  real<lower=0> lam_sig;    // lambda for the signal where there is no peak
  real<lower=0,upper=1> pa; // proportion of allowed distribution below 0
}

transformed data{
    vector[n] lam;       // lambda at each timepoint
    real<lower=0> kappa; // calculated kappa from pa
    kappa=sqrt(pa)/sqrt(1-pa);
    
    lam=rep_vector(lam_sig, n); 
    for(i in 1:npeaks){
        lam[peakix[i]]=lam_prominences[i];
    }
}
parameters {
    vector[ncol] coef; // coefficients for the basis-functions
}

transformed parameters{
    
}

model {
    {
    vector[n] d;
        
    coef ~ normal(0,5);
    d=sy-(B*coef); // center at estimated baseline
    for( i in 1:n ){
        d[i] ~ skew_double_exponential2(lam[i], kappa);
    }
    }
}
"""


def bspline(txd, knots, spline_degree=3):
    """
    Re-implementation from https://mc-stan.org/users/documentation/case-studies/splines_in_stan.html.
    Similar behaviour as R's bs() function from the splines-library.
    
    Parameters
    -----------
    
    txd: np.array
        time-vector
    knots: np.array
        location of the knots
    spline_degree: int
        degree of the spline
        
    Returns
    --------
    
    B: np.array
        matrix of basis functions
    """
    n=txd.shape[0]
    num_knots=knots.shape[0]

    def build_b_spline(t, ext_knots, ind, order):
        n=t.shape[0]
        b_spline=np.zeros(n)
        w1=np.zeros(n)
        w2=np.zeros(n)
        if order==1:
            b_spline=np.zeros(n)
            b_spline[np.logical_and(t>=ext_knots[ind], t<ext_knots[ind+1]) ]=1
        else:
            if ext_knots[ind] != ext_knots[ind+order-1]:
                w1=(t-np.array([ext_knots[ind]]*n))/(ext_knots[ind+order-1]-ext_knots[ind])
            if ext_knots[ind+1]!=ext_knots[ind+order]:
                w2=1-(t-np.array([ext_knots[ind+1]]*n))/(ext_knots[ind+order]-ext_knots[ind+1])
            b_spline=w1*build_b_spline(t,ext_knots,ind,order-1)+w2*build_b_spline(t,ext_knots,ind+1,order-1)
        return b_spline

    num_basis = num_knots + spline_degree - 1; # total number of B-splines
    B=np.zeros( (num_basis, n) )
    ext_knots=np.concatenate( ([knots[0]]*spline_degree, knots, 
                              [knots[num_knots-1]]*spline_degree) )
    for i in range(num_basis):
        B[i,:] = build_b_spline(txd, ext_knots, i, spline_degree+1)
    B[num_basis-1,n-1]=1
    return B.T

def butter_lowpass(cutoff, fs, order=5):
    """
    Get lowpass-filter coefficients for Butterworth-filter.
    
    Parameters
    -----------
    
    cutoff: float
        lowpass-filter cutoff
    fs: float
        sampling rate
    order: int
        filter order
        
    Returns
    -------
    
    (b,a): tuple (float,float)
        filter coefficients
    """
    nyq = 0.5 * fs
    normal_cutoff = cutoff / nyq
    b, a = signal.butter(order, normal_cutoff, btype='low', analog=False)
    return b, a

def butter_lowpass_filter(data, cutoff, fs, order=5):
    """
    Lowpass-filter signal using a Butterworth-filter.
    
    Parameters
    -----------
    
    data: np.array
        data to lowpass-filter
    cutoff: float
        lowpass-filter cutoff
    fs: float
        sampling rate
    order: int
        filter order
        
    Returns
    -------
    
    y: np.array
        filtered data
    """
    b, a = butter_lowpass(cutoff, fs, order=order)
    y = signal.filtfilt(b, a, data)
    return y

def downsample(y,R):
    """
    Simple downsampling scheme using mean within the downsampling window.
    
    Parameters
    -----------
    
    y: np.array
        signal to downsample
        
    R: int
        decimate-factor
        
    Returns
    -------
    
    y: np.array
        downsampled data
    """
    pad_size = int(math.ceil(float(y.size)/R)*R - y.size)
    y_padded = np.append(y, np.zeros(pad_size)*np.NaN)
    y2=scipy.nanmean(y_padded.reshape(-1,R), axis=1)
    return y2


    
def baseline_envelope_iter_bspline(tx,sy,event_onsets,fs, fsd=10, lp=2, 
                                   lam_sig=1, lam_min=1, lam_max=100,
                                   verbose=0):
    """
    Extract baseline based (re-)estimating the lower envelope using B-splines.
    See notebook `baseline_interp.ipynb` for details.
    The signal is downsampled (to `fsd` Hz) for speed.
    
    Parameters
    -----------
    
    tx : np.ndarray
        time-vector in seconds
        
    sy : np.ndarray
        raw pupil signal
        
    event_onsets : list
        onsets of events (stimuli/responses) in milliseconds
        
    fs : float
        sampling rate in Hz
    
    fsd : float
        downsampled sampling rate (if too slow, decrease)
       
    lp : float
        lowpass-filter cutoff (Hz)
    
    lam_sig: float
        parameter steering how much the baseline is shaped by the non-peaks of the signal
    
    lam_min,lam_max: float
        parameters mapping how much low- and high-prominence peaks influence the baseline
        
    verbose: int [0, 100]
        how much information to print (0 nothing, 100 everything)
    
    Returns
    -------
    
    (txd,syd,base2,base1) : tuple
        txd: downsampled time-array
        syd: downsampled and lowpass-filtered pupil signal
        base1: is the estimated base after the first round
        base2: is the final baseline estimate
        
    """
    def vprint(v, s):
        if v<=verbose:
            print(">",s)

    dsfac=int(fs/fsd) # calculate downsampling factor
    vprint(100, "Downsampling factor is %i"%dsfac)
    
    # downsampling
    syc=butter_lowpass_filter(sy, lp, fs, order=2)
    syd=downsample(syc, dsfac)

    # z-scale for easier model-fitting
    symean,sysd=np.mean(syd),np.std(syd)
    syd=(syd-symean)/sysd
    txd=downsample(tx, dsfac)
    vprint(100, "Downsampling done")

    # peak finding and spline-building
    peaks_ix=signal.find_peaks(-syd)[0]
    prominences=signal.peak_prominences(-syd, peaks_ix)[0]
    peaks=txd[peaks_ix]
    vprint(100, "Peak-detection done, %i peaks detected"%peaks.shape[0])
    knots=np.concatenate( ([txd.min()], peaks, [txd.max()]) ) ## peaks as knots
    B=bspline(txd, knots, 3)
    vprint(100, "B-spline matrix built, dims=%s"%str(B.shape))
    

    # convert
    def prominence_to_lambda(w, lam_min=1, lam_max=100):
        w2=lam_min+((w-np.min(w))/(np.max(w-np.min(w))))*(lam_max-lam_min)
        return w2
    
    w=prominence_to_lambda(prominences, lam_min=lam_min, lam_max=lam_max)
    
    # load or compile model
    vprint(10, "Compiling Stan model")

    sm = StanModel_cache(stan_code_baseline_model_asym_laplac)
    
    ## put the data for the model together
    data={
        'n':syd.shape[0],
        'sy':syd,
        'ncol':B.shape[1],
        'B':B,
        'npeaks':peaks_ix.shape[0],
        'peakix':peaks_ix,
        'lam_sig':lam_sig,
        'pa':0.05,
        'lam_prominences':w
    }
    
    ## variational optimization
    vprint(10, "Optimizing Stan model")
    opt=sm.vb(data=data)
    vbc=opt["mean_pars"]
    meansigvb=np.dot(B, vbc)
    vprint(10, "Done optimizing Stan model")
    
    ## PRF model
    # new "signal"
    syd2=syd-meansigvb

    vprint(10, "Estimating PRF model (NNLS)")    
    #coef,pred,resid=pupilresponse_nnls(txd,syd2,event_onsets,fs=fsd)
    pred, coef, _, _, _=pupil_response(txd, syd2, event_onsets, fsd, npar=10, tmax=917)
    resid=syd-pred
    vprint(10, "Done Estimating PRF model (NNLS)")
    
    ### 2nd iteration
    ## get new peaks
    syd3=syd-pred
    peaks2_ix=signal.find_peaks(-syd3)[0]
    prominences2=signal.peak_prominences(-syd3, peaks2_ix)[0]
    peaks2=txd[peaks2_ix]
    vprint(100, "2nd Peak-detection done, %i peaks detected"%peaks2.shape[0])

    knots2=np.concatenate( ([txd.min()], peaks2, [txd.max()]) ) ## peaks as knots
    B2=bspline(txd, knots2, 3)
    vprint(100, "2nd B-spline matrix built, dims=%s"%str(B2.shape))

    w2=prominence_to_lambda(prominences2, lam_min=lam_min, lam_max=lam_max)

    data2={
        'n':syd3.shape[0],
        'sy':syd3,
        'ncol':B2.shape[1],
        'B':B2,
        'npeaks':peaks2_ix.shape[0],
        'peakix':peaks2_ix,
        'lam_sig':lam_sig,
        'pa':0.05,
        'lam_prominences':w2
    }
    
    ##  variational optimization
    vprint(10, "2nd Optimizing Stan model")
    opt=sm.vb(data=data2)
    vbc2=opt["mean_pars"]
    meansigvb2=np.dot(B2, vbc2)  
    vprint(10, "Done 2nd Optimizing Stan model")
    
    return txd,(syd*sysd)+symean,(meansigvb2*sysd)+symean, (meansigvb*sysd)+symean

def baseline_envelope(tx,sy,event_onsets, fs=1000, lp=2, prominence_thr=80, interp_method="cubic"):
    """
    Extract baseline based on the lower envelope of the (filtered) signal.

    Steps: 
    
    - filter away noise
    - detect high-prominence throughs in the signal 
    - calculate lower envelope based on these peaks
    
    Parameters
    -----------
    
    tx : np.ndarray
        time-vector in seconds
        
    sy : np.ndarray
        raw pupil signal
        
    event_onsets : list
        onsets of events (stimuli/responses) in seconds
        
    fs : float
        sampling rate in Hz
    
    lp : float
        low-pass filter cutoff for removing random noise
        
    prominence_thr : float in [0,100]
        percentile of the prominence distribution (of the peaks) to 
        use for determining prominent peaks (see `scipy.stats.peak_prominences()`)
        
    interp_method : string, one of ["linear", "cubic", "spline"]
        "linear" - linear interpolation between the high-prominence peaks
        "cubic"  - cubic interpolation through all high-prominence peaks
        "spline" - a smoothing spline that is guaranteed to go through all
                   high-prominence peaks and smoothes through all the other
                   (lower-prominence) peaks
                   
    Returns
    --------
    
    base: np.array
        baseline estimate
    """
    syc=butter_lowpass_filter(sy, fs=fs, order=2, cutoff=lp)
    peaks_ix=signal.find_peaks(-syc)[0]
    prominences=signal.peak_prominences(-syc, peaks_ix)[0]
    res=signal.peak_widths(-syc, peaks_ix)
    width_locs=(-res[1],res[2]/fs,res[3]/fs)
    widths=res[0]
    peaks=tx[peaks_ix]
    widths=widths/fs # in seconds
    prominence_cutoff=np.percentile(prominences,prominence_thr)
    real_peaks=peaks[prominences>prominence_cutoff]
    real_peaks_ix=peaks_ix[prominences>prominence_cutoff]
    
    if interp_method in ["linear","cubic"]:
        ## interpolate only most prominent peaks
        xinterp=np.concatenate( ([tx.min()],real_peaks,[tx.max()]) )
        yinterp=np.concatenate( ([syc[0]], syc[real_peaks_ix], [syc[-1]]) )
        f=interp1d(xinterp,yinterp, kind=interp_method)
    elif interp_method=="spline":
        ## use all peaks for interpolation and use "real" peaks as inner knots
        xinterp=np.concatenate( ([tx.min()],peaks,[tx.max()]) )
        yinterp=np.concatenate( ([syc[0]], syc[peaks_ix], [syc[-1]]) )
        f=scipy.interpolate.LSQUnivariateSpline(xinterp,yinterp,real_peaks)
    else:
        raise ValueError("interp_method must be one of 'linear','cubic','spline'")
    x0=f(tx)
    
    return x0


def baseline_pupil_model(tx,sy,event_onsets, fs=1000, lp1=2, lp2=0.2):
    """
    Extract baseline based on filtering after removing stim-locked activity.
    
    Steps:
    
    - filter away noise
    - regress out event-locked activity from the filtered signal using NNLS
    - remove modeled signal from filtered data
    - run another lowpass-filter to get rid of spurious signals
    
    Parameters
    -----------
    
    tx : np.ndarray
        time-vector in seconds
        
    sy : np.ndarray
        raw pupil signal
        
    event_onsets : list
        onsets of events (stimuli/responses) in seconds
        
    fs : float
        sampling rate in Hz
    
    lp1 : float
        low-pass filter cutoff for removing random noise
        
    lp2 : float
        low-pass filter cutoff for removing spurious peaks from the baseline-signal        
        
    Returns
    --------
    
    base: np.array
        baseline estimate        
    """
    syc=butter_lowpass_filter(sy, fs=fs, order=2, cutoff=lp1)
    
    # calculate indices for event-onsets
    event_onsets_ix=np.argmin(np.abs(np.tile(event_onsets, (sy.size,1)).T-tx), axis=1)

    # set up a single regressor
    x1=np.zeros(sy.size, dtype=np.float)
    x1[event_onsets_ix]=1
    kernel=pupil_kernel(4, fs=fs)
    x1=np.convolve(x1, kernel, mode="full")[0:x1.size]

    # solve with non-negative least-squares
    X=np.stack( (x1, np.ones(x1.size)))
    coef=scipy.optimize.nnls(X.T, syc)[0]
    pred=coef[1]+coef[0]*x1
    resid=syc-pred+coef[1]

    resid_lp=butter_lowpass_filter(resid, fs=fs, order=2, cutoff=lp2)
    x0 = resid_lp
    
    return x0
